from typing import Optional, List
from ...data import DomainEvent
from .types import ConnectionType, LoadNetworkResourcePageResult, Cookie


class Network:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Network
    """
    __slots__ = ("_connection", "enabled")

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn
        self.enabled   = False


    async def enable(
            self,
            maxTotalBufferSize:    Optional[int] = None,
            maxResourceBufferSize: Optional[int] = None,
            maxPostDataSize:       Optional[int] = None
    ) -> None:
        """
        Включает отслеживание сети, сетевые события теперь будут доставляться клиенту.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-enable
        :param maxTotalBufferSize:      (optional, EXPERIMENTAL) Размер буфера в байтах для использования
                                            при сохранении полезных данных сети (XHR и т. Д.).
        :param maxResourceBufferSize:   (optional, EXPERIMENTAL) Размер буфера для каждого ресурса в
                                            байтах для использования при сохранении полезных данных сети
                                            (XHR и т. Д.).
        :param maxPostDataSize:         (optional) Самый длинный размер тела сообщения (в байтах),
                                            который будет включен в уведомление "requestWillBeSent".
        :return:
        """
        if not self.enabled:
            args = {}
            if maxTotalBufferSize is not None: args.update({"maxTotalBufferSize": maxTotalBufferSize})
            if maxResourceBufferSize is not None: args.update({"maxResourceBufferSize": maxResourceBufferSize})
            if maxPostDataSize is not None: args.update({"maxPostDataSize": maxPostDataSize})
            await self._connection.call("Network.enable", args)
            self.enabled = True

    async def disable(self) -> None:
        """
        Отключает отслеживание сети, запрещает отправку сетевых событий клиенту.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-disable
        :return:
        """
        if self.enabled:
            await self._connection.call("Network.disable")
            self.enabled = False

    async def emulateNetworkConditions(
        self, latency: int,
        downloadThroughput: int = -1,
        uploadThroughput:   int = -1,
        offline:           bool = False,
        connectionType: Optional[ConnectionType] = None
    ) -> None:
        """
        Активирует эмуляцию состояния сети.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-emulateNetworkConditions
        :param latency:             Минимальная задержка от запроса, отправленного полученным заголовкам
                                        ответа (мс).
        :param downloadThroughput:  (optional) Максимальная агрегированная скорость скачивания (байт / с).
                                        -1 отключает регулирование.
        :param uploadThroughput:    (optional) Максимальная агрегированная скорость загрузки (байт / с).
                                        -1 отключает регулирование.
        :param offline:             (optional) 'True' — эмулирует отключение от интернета.
        :param connectionType:      (optional) Основная технология подключения, которую, предположительно
                                        использует браузер.
                                        Allowed values: none, cellular2g, cellular3g, cellular4g,
                                        bluetooth, ethernet, wifi, wimax, other
        :return:
        """
        args = {"latency": latency, "offline": offline}
        if downloadThroughput > -1: args.update({"downloadThroughput": downloadThroughput})
        if uploadThroughput > -1: args.update({"uploadThroughput": uploadThroughput})
        if connectionType: args.update({"connectionType": connectionType.value})
        await self._connection.call("Network.emulateNetworkConditions", args)

    async def clearBrowserCache(self) -> None:
        """
        Clears browser cache.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-clearBrowserCache
        :return:
        """
        await self._connection.call("Network.clearBrowserCache")

    async def clearBrowserCookies(self) -> None:
        """
        Clears browser cookies.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-clearBrowserCookies
        :return:
        """
        await self._connection.call("Network.clearBrowserCookies")

    async def setBlockedURLs(self, urls: List[str]) -> None:
        """
        (EXPERIMENTAL)
        Блокирует загрузку URL-адресов.
        !!!ВНИМАНИЕ!!! Требует активации доменов Page и Network!
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-setBlockedURLs
        :param urls:            Шаблоны URL для блокировки. Подстановочные знаки ('*') разрешены.
        :return:
        """
        if not self.enabled:
            await self.enable()
        await self._connection.call("Network.setBlockedURLs", {"urls": urls})

    async def setCacheDisabled(self, cacheDisabled: bool = True) -> None:
        """
        Включает игнорирование кеша для каждого запроса. Если 'true', кеш не будет использоваться.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-setCacheDisabled
        :param cacheDisabled:    Состояние.
        :return:
        """
        await self._connection.call("Network.setCacheDisabled", {"cacheDisabled": cacheDisabled})

    async def getAllCookies(self) -> List[Cookie]:
        """
        Возвращает все куки браузера. В зависимости от поддержки бэкэнда, вернет подробную
            информацию о куки в поле куки.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-getAllCookies
        :return: cookies -> (array Cookie) Array of cookie objects.
        """
        cookies = []
        for c in (await self._connection.call("Network.getAllCookies"))["cookies"]:
            cookies.append(Cookie(**c))
        return cookies

    async def getCookies(self, urls: Optional[list] = None) -> List[Cookie]:
        """
        Возвращает все куки браузера для текущего URL. В зависимости от поддержки бэкэнда,
            вернет подробную информацию о куки в поле куки.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-getCookies
        :param urls: список строк содержащих адреса, для которых будут извлечены Cookies [ "https://google.com", ... ]
        :return: cookies -> (array Cookie) Array of cookie objects.
        """
        args = {}
        if urls: args.update({"urls": urls})
        cookies = []
        for c in (await self._connection.call("Network.getCookies", args))["cookies"]:
            cookies.append(Cookie(**c))
        return cookies

    async def deleteCookies(
            self, name: str,
            url:    str = "",
            domain: str = "",
            path:   str = ""
    ) -> None:
        """
        Удаляет файлы cookie браузера с соответствующими именами и URL-адресами или парой домен / путь.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-deleteCookies
        :param name:    Имя куки для удаления.
        :param url:     (optional) Если указан, удаляет все куки с указанным именем, где 'domain' и
                            'path' соответствуют указанному URL.
        :param domain:  (optional) Если указан, удаляет только те куки, что точно соответствуют 'domain'.
        :param path:    (optional) Если указан, удаляет только те куки, что точно соответствуют 'path'.
        :return:
        """
        args = {"name": name}
        if url: args.update({"url": url})
        if domain: args.update({"domain": domain})
        if path: args.update({"path": path})
        await self._connection.call("Network.deleteCookies", args)

    async def setCookie(
            self, name: str, value: str,
            url:       str = "",
            domain:    str = "",
            path:      str = "",
            secure:   Optional[bool] = None,
            httpOnly: Optional[bool] = None,
            sameSite:  str = "",
            expires:   int = -1,
            priority:  str = ""
    ) -> bool:
        """
        Устанавливает cookie с указанными данными cookie. Если они существуют, то будут перезаписаны.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-setCookie
        :param name:        Cookie name.
        :param value:       Cookie value.
        :param url:         (optional) The request-URI to associate with the setting of
                                the cookie. This value can affect the default domain and
                                path values of the created cookie.
        :param domain:      (optional) Cookie domain.
        :param path:        (optional) Cookie path.
        :param secure:      (optional) True if cookie is secure.
        :param httpOnly:    (optional) True if cookie is http-only.
        :param sameSite:    (optional) Cookie SameSite type. Represents the cookie's 'SameSite'
                                status: https://tools.ietf.org/html/draft-west-first-party-cookies
                                Allowed values: Strict, Lax, None
        :param expires:     (optional) Cookie expiration date, session cookie if not set.
                                UTC time in seconds, counted from January 1, 1970.
        :param priority:    (optional, EXPERIMENTAL) Cookie Priority type. Represents the cookie's 'Priority'
                                status: https://tools.ietf.org/html/draft-west-cookie-priority-00
                                Allowed values: Low, Medium, High
        :return:            True if successfully set cookie.
        """
        args = {"name": name, "value": value}
        if url: args.update({"url": url})
        if domain: args.update({"domain": domain})
        if path: args.update({"path": path})
        if secure is not None: args.update({"secure": secure})
        if secure is not None: args.update({"httpOnly": httpOnly})
        if sameSite: args.update({"sameSite": sameSite})
        if expires > -1: args.update({"expires": expires})
        if priority: args.update({"priority": priority})
        return (await self._connection.call("Network.setCookie", args))["success"]

    async def setCookies(self, list_cookies: list) -> None:
        """
        Устанавливает сразу список кук
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-setCookies
        :param list_cookies:        список куки-параметров
        :return:
        """
        await self._connection.call("Network.setCookies", {"cookies": list_cookies})

    async def setExtraHTTPHeaders(self, headers: dict) -> None:
        """
        Устанавливает дополнительные заголовки, которые всегда будут отправляться в запросах
            от инстанса текущей страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-setExtraHTTPHeaders
        :param headers:        Заголовки запроса / ответа в виде ключей / значений объекта JSON.
        :return:
        """
        await self._connection.call("Network.setExtraHTTPHeaders", {"headers": headers})

    async def setUserAgentOverride(
        self, userAgent: str,
        acceptLanguage:     Optional[str] = None,
        platform:           Optional[str] = None,
        userAgentMetadata: Optional[dict] = None
    ) -> None:
        """
        Позволяет переопределить пользовательский агент с заданной строкой. Функционал ничем не
            отличается от одноимённого метода домена 'Emulation'.
        https://chromedevtools.github.io/devtools-protocol/tot/Network/#method-setUserAgentOverride
        :param userAgent:           Новый юзер-агент.
        :param acceptLanguage:      (optional) Язык браузера для эмуляции.
        :param platform:            (optional) Платформа браузера, которую возвращает
                                        "navigator.platform".
                                        https://www.w3schools.com/jsref/prop_nav_platform.asp
        :param userAgentMetadata:   (optional, EXPERIMENTAL) Для отправки в заголовках Sec-CH-UA- * и возврата в
                                        navigator.userAgentData. Ожидатся словарь вида:
                                        {
                                            "brands": [{"brand": "brand name", "version": "brand version"}, { ... }, ... ],
                                            "fullVersion": "full version",
                                            "platform": "platform name",
                                            "platformVersion": "platform version",
                                            "architecture": "devise architecture",
                                            "model": "model",
                                            "mobile": boolean,
                                        }
        :return:            None
        """
        args = {"userAgent": userAgent}
        if acceptLanguage: args.update({"acceptLanguage": acceptLanguage})
        if platform: args.update({"platform": platform})
        if userAgentMetadata: args.update({"userAgentMetadata": userAgentMetadata})
        await self._connection.call("Network.setUserAgentOverride", args)

    async def loadNetworkResource(
        self,
        url:      Optional[str] = None,
        options: Optional[dict] = None,
        frameId:  Optional[str] = None
    ) -> LoadNetworkResourcePageResult:
        """
        Выбирает ресурс и возвращает контент.
        https://chromedevtools.github.io/devtools-protocol/tot/Network#method-loadNetworkResource
        :param url:             (optional) URL ресурса, для которого нужно получить контент.
        :param options:         (optional) Опции запроса
        :param frameId:         (optional) Идентификатор фрейма
        :return:
        """
        if url is None: url = await self._connection.extend.getUrl()
        if options is None: options = {"disableCache": False, "includeCredentials": True}
        if frameId is None: frameId = self._connection.conn_id
        args = { "url": url, "options": options, "frameId": frameId }
        resource = (await self._connection.call("Network.loadNetworkResource", args))["resource"]
        return LoadNetworkResourcePageResult(**resource)


class NetworkEvent(DomainEvent):
    dataReceived = "Network.dataReceived"
    eventSourceMessageReceived = "Network.eventSourceMessageReceived"
    loadingFailed = "Network.loadingFailed"
    loadingFinished = "Network.loadingFinished"
    requestServedFromCache = "Network.requestServedFromCache"
    requestWillBeSent = "Network.requestWillBeSent"
    responseReceived = "Network.responseReceived"
    webSocketClosed = "Network.webSocketClosed"
    webSocketCreated = "Network.webSocketCreated"
    webSocketFrameError = "Network.webSocketFrameError"
    webSocketFrameReceived = "Network.webSocketFrameReceived"
    webSocketFrameSent = "Network.webSocketFrameSent"
    webSocketHandshakeResponseReceived = "Network.webSocketHandshakeResponseReceived"
    webSocketWillSendHandshakeRequest = "Network.webSocketWillSendHandshakeRequest"
    webTransportClosed = "Network.webTransportClosed"
    webTransportConnectionEstablished = "Network.webTransportConnectionEstablished"
    webTransportCreated = "Network.webTransportCreated"
    reportingApiEndpointsChangedForOrigin = "Network.reportingApiEndpointsChangedForOrigin"
    reportingApiReportAdded = "Network.reportingApiReportAdded"
    reportingApiReportUpdated = "Network.reportingApiReportUpdated"
    requestWillBeSentExtraInfo = "Network.requestWillBeSentExtraInfo"
    resourceChangedPriority = "Network.resourceChangedPriority"
    responseReceivedExtraInfo = "Network.responseReceivedExtraInfo"
    signedExchangeReceived = "Network.signedExchangeReceived"
    subresourceWebBundleInnerResponseError = "Network.subresourceWebBundleInnerResponseError"
    subresourceWebBundleInnerResponseParsed = "Network.subresourceWebBundleInnerResponseParsed"
    subresourceWebBundleMetadataError = "Network.subresourceWebBundleMetadataError"
    subresourceWebBundleMetadataReceived = "Network.subresourceWebBundleMetadataReceived"
    trustTokenOperationDone = "Network.trustTokenOperationDone"
    requestIntercepted = "Network.requestIntercepted"





class DeviceOrientation:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/DeviceOrientation
    """
    __slots__ = ("_connection",)

    def __init__(self, conn) -> None:
        from ...connection import Connection

        self._connection: Connection = conn

    async def clearDeviceOrientationOverride(self) -> None:
        """
        Очищает переопределенную ориентацию устройства.
        https://chromedevtools.github.io/devtools-protocol/tot/DeviceOrientation/#method-clearDeviceOrientationOverride
        :return:
        """
        await self._connection.call("DeviceOrientation.clearDeviceOrientationOverride")

    async def setDeviceOrientationOverride(self, alpha: float, beta: float, gamma: float) -> None:
        """
        Переопределяет ориентацию устройства, принудительно изменяя значения сенсоров, котоые так же
            можно найти в консоли браузера по Ctrl+Shift+P и в поиске ввести 'Show Sensors'.
        https://chromedevtools.github.io/devtools-protocol/tot/DeviceOrientation/#method-setDeviceOrientationOverride
        :return:
        """
        args = {"alpha": alpha, "beta": beta, "gamma": gamma}
        await self._connection.call("DeviceOrientation.setDeviceOrientationOverride", args)



from typing import Optional, List
from ...data import DomainEvent
from .types import Version, Bounds, WindowInfo


class Browser:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Browser/
    """
    __slots__ = ("_connection",)

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn

    async def setPermission(
            self, permission: dict, setting: str, origin: str = None,
            browserContextId: str = None
    ) -> None:
        """
        (EXPERIMENTAL)
        Устанавливает настройки разрешений для данного источника.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-setPermission
        :param permission:          Дескриптор разрешения для переопределения, определяется
                                        словарём, с обязательным атрибутом "name". Подробней:
                                            {
                                                "name": str(),             -> Имя разрешения.
                                                            https://cs.chromium.org/chromium/src/third_party/blink/renderer/modules/permissions/permission_descriptor.idl
                                                "sysex": bool(),           -> (optional) Для разрешения
                                                            «midi» может также указываться sysex control.
                                                "userVisibleOnly": bool(), -> (optional) Для разрешения
                                                            «push» можно указать userVisibleOnly. Обратите
                                                            внимание, что userVisibleOnly = true -
                                                            единственный поддерживаемый в настоящее время тип.
                                                "type": str(),             -> (optional) Для разрешения
                                                            "wake-lock" необходимо указать тип "screen"
                                                            или "system".
                                                "allowWithoutSanitization": bool() -> (optional) Для
                                                            разрешения "clipboard" можно указать
                                                            allowWithoutSanitization.
                                            }
        :param setting:             Настройки разрешения. Могут быть: granted, denied, prompt
        :param origin:              (optional) Источник, к которому относится разрешение. Если не указано,
                                        подразумеваются все.
        :param browserContextId:    (optional) Контекст для переопределения. Если не указано, используется
                                        контекст браузера по умолчанию.
        :return:
        """
        args = {"permission": permission, "setting": setting}
        if origin is not None:
            args.update({"origin": origin})
        if browserContextId is not None:
            args.update({"browserContextId": browserContextId})
        await self._connection.call("Browser.setPermission", args)

    async def grantPermissions(
            self, permissions: List[str],
            origin: Optional[str] = None,
            browserContextId: Optional[str] = None
    ) -> None:
        """
        (EXPERIMENTAL)
        Предоставляет определенные разрешения данному источнику, отклоняя все остальные.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-grantPermissions
        :param permissions:         Возможные значения:
                                        accessibilityEvents, audioCapture, backgroundSync, backgroundFetch,
                                        clipboardReadWrite, clipboardSanitizedWrite, durableStorage, flash,
                                        geolocation, midi, midiSysex, nfc, notifications, paymentHandler,
                                        periodicBackgroundSync, protectedMediaIdentifier, sensors, videoCapture,
                                        idleDetection, wakeLockScreen, wakeLockSystem
        :param origin:              (optional) Источник, к которому относится разрешение. Если не указано,
                                        подразумеваются все.
        :param browserContextId:    (optional) Контекст для переопределения. Если не указано, используется
                                        контекст браузера по умолчанию.
        :return:
        """
        args = {"permissions": permissions}
        if origin is not None:
            args.update({"origin": origin})
        if browserContextId is not None:
            args.update({"browserContextId": browserContextId})
        await self._connection.call("Browser.grantPermissions", args)

    async def resetPermissions(self, browserContextId: Optional[str] = None) -> None:
        """
        Сбросить все управление разрешениями для всех источников.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-resetPermissions
        :param browserContextId:    (optional) Контекст для переопределения. Если не указано, используется
                                        контекст браузера по умолчанию.
        :return:
        """
        args = {}
        if browserContextId is not None:
            args.update({"browserContextId": browserContextId})
        await self._connection.call("Browser.resetPermissions", args)

    async def getVersion(self) -> Version:
        """
        Возвращает словарь с информацией о текущем билде браузера.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-getVersion
        :return:                {
                                    "protocolVersion":  str( ... ), -> Protocol version.
                                    "product":          str( ... ), -> Product name.
                                    "revision":         str( ... ), -> Product revision.
                                    "userAgent":        str( ... ), -> User-Agent.
                                    "jsVersion":        str( ... )  -> V8 version.
                                }
        """
        return Version(
            **(await self._connection.call("Browser.getVersion")))

    async def getWindowBounds(self, windowId: int = None) -> Bounds:
        """
        (EXPERIMENTAL)
        Возвращает позицию и размер окна.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-getWindowBounds
        :param windowId:        Идентификатор окна браузера.
        :return:                {
                                    "left":       int(), -> (optional) Смещение от левого
                                                                края экрана до окна в пикселях.
                                    "top":        int(), -> (optional) Смещение от верхнего
                                                                края экрана до окна в пикселях.
                                    "width":      int(), -> (optional) Ширина окна в пикселях.
                                    "height":     int(), -> (optional) Высота окна в пикселях
                                    "windowState": str() -> (optional) normal, minimized,
                                                                maximized, fullscreen
                                }
        """
        if windowId is None:
            windowId = (await self._connection.Target.getWindowForTarget()).windowId
        return Bounds(
            **(await self._connection.call("Browser.getWindowBounds", {"windowId": windowId}))["bounds"])

    async def getWindowForTarget(self, targetId: Optional[str] = None) -> WindowInfo:
        """
        (EXPERIMENTAL)
        Возвращает идентификатор, а так же позицию и размер окна.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-getWindowForTarget
        :param targetId:        (optional) Идентификатор хоста агента Devtools. Если вызывается
                                    как часть сеанса, используется связанный targetId.
        :return:                {
                                    "windowId": int(), -> Идентификатор окна.
                                    "bounds":   dict() -> То же, что возвращает GetWindowBounds().
                                }
        """
        if targetId is None:
            targetId = self._connection.conn_id
        result = await self._connection.call("Browser.getWindowForTarget", {"targetId": targetId})
        return WindowInfo(windowId=result["windowId"], bounds=Bounds(**result["bounds"]))

    async def setDockTile(self, badgeLabel: Optional[str] = None, image: Optional[str] = None) -> None:
        """
        (EXPERIMENTAL)
        Задать сведения о док-плитке для конкретной платформы.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-setDockTile
        :param badgeLabel:      (optional) Значок метки(?)
        :param image:           (optional) PNG кодированное изображение.
        :return:
        """
        args = {}
        if badgeLabel is not None:
            args.update({"badgeLabel": badgeLabel})
        if image is not None:
            args.update({"image": image})
        if not args:
            return
        await self._connection.call("Browser.setDockTile", args)

    async def setDownloadBehavior(
            self, behavior: str,
            browserContextId: Optional[str] = None,
            downloadPath: Optional[str] = None,
            eventsEnabled: Optional[bool] = None
    ) -> None:
        """
        Устанавливает поведение при загрузке файлов.
        https://chromedevtools.github.io/devtools-protocol/tot/Page#method-setDownloadBehavior
        :param behavior:            Разрешить все или отклонить все запросы на загрузку, или использовать поведение
                                        Chrome по умолчанию, если доступно (в противном случае запретить).
                                        deny = запрет, allow = разрешить, default
        :param browserContextId:    (optional) Если не указано, используется контекст браузера по умолчанию.
        :param downloadPath:        (optional) Путь по умолчанию для сохранения загруженных файлов. Это необходимо,
                                        если для поведения установлено значение 'allow' и если путь не передан, будет
                                        установлено текущее расположение.
        :param eventsEnabled:       (optional) Если не указано, используется контекст браузера по умолчанию.
        :return:
        """
        args = {"behavior": behavior}
        if browserContextId is not None: args.update(browserContextId=browserContextId)
        if downloadPath is not None: args.update(downloadPath=downloadPath)
        if eventsEnabled is not None: args.update(eventsEnabled=eventsEnabled)
        await self._connection.call("Browser.setDownloadBehavior", args)

    async def close(self) -> bool:
        """
        Изящно завершает работу браузера.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-close
        :return:        Закрылся/был закрыт
        """
        if self._connection.connected:
            await self._connection.call("Browser.close")
            return True
        return False


class BrowserEvent(DomainEvent):
    downloadProgress = "Browser.downloadProgress"
    downloadWillBegin = "Browser.downloadWillBegin"



from dataclasses import dataclass
from typing import Optional


@dataclass
class StyleProp:
    """ Описывает имя и значение стиля """
    name: str
    value: str


@dataclass
class NodeCenter:
    x: int
    y: int



@dataclass
class NodeRect:
    """ Пространственное положение и размеры узла """
    x: int
    y: int
    width: int
    height: int
    left: int
    right: int
    top: int
    bottom: int



@dataclass
class ShapeOutsideInfo:
    bounds: list
    shape: list
    marginShape: list


@dataclass
class BoxModel:
    content: list
    padding: list
    border: list
    margin: list
    width: int
    height: int
    shapeOutside: Optional[ShapeOutsideInfo] = None



from ...data import DomainEvent


class Overlay:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Overlay
    """
    __slots__ = ("_connection", "enabled")

    def __init__(self, conn) -> None:
        from ...connection import Connection

        self._connection: Connection = conn
        self.enabled = False

    async def enable(self) -> None:
        """
        Включает отправку уведомлений домена 'Overlay'.
        https://chromedevtools.github.io/devtools-protocol/tot/Overlay#method-enable
        :return:
        """
        if not self.enabled:
            await self._connection.call("Overlay.enable")
            self.enabled = True

    async def disable(self) -> None:
        """
        Выключает 'Overlay' домен, останавливая отправку сообщений.
        https://chromedevtools.github.io/devtools-protocol/tot/Overlay#method-disable
        :return:
        """
        if self.enabled:
            await self._connection.call("Overlay.disable")
            self.enabled = False

    async def setShowFPSCounter(self) -> None:
        """
        Запрашивает показ счетчика FPS.
        https://chromedevtools.github.io/devtools-protocol/tot/Overlay#method-setShowFPSCounter
        :return:
        """
        await self._connection.call("Overlay.setShowFPSCounter")


class OverlayEvent(DomainEvent):
    inspectModeCanceled = "Overlay.inspectModeCanceled"
    inspectNodeRequested = "Overlay.inspectNodeRequested"
    nodeHighlightRequested = "Overlay.nodeHighlightRequested"
    screenshotRequested = "Overlay.screenshotRequested"



from dataclasses import dataclass, field
from typing import Literal, Optional, List, Union, Dict


@dataclass
class BindingCalledData:
    name: str
    payload: str
    executionContextId: int


@dataclass
class SerializationOptions:
    serialization: Literal["deep", "json", "idOnly"]
    maxDepth: Optional[int] = None

    def as_dict(self) -> Dict[str, Union[str, int]]:
        result = {"serialization": self.serialization}
        if self.maxDepth is not None:
            result.update({"maxDepth": self.maxDepth})
        return result


@dataclass
class PropertyPreview:
    name: str
    type: Literal["object", "function", "undefined", "string", "number", "boolean", "symbol", "accessor", "bigint"]
    valuePreview: Optional['ObjectPreview']
    value: Optional[str] = None
    _valuePreview: Optional['ObjectPreview'] = field(init=False, repr=False, default=None)
    subtype: Literal[
        "array", "null", "node", "regexp", "date", "map", "set", "weakmap", "weakset", "iterator", "generator",
        "error", "proxy", "promise", "typedarray", "arraybuffer", "dataview", "webassemblymemory", "wasmvalue"] = None

    @property
    def valuePreview(self) -> 'ObjectPreview':
        return self._valuePreview

    @valuePreview.setter
    def valuePreview(self, data: dict) -> None:
        self._valuePreview = ObjectPreview(**data) if not isinstance(data, property) else None


@dataclass
class EntryPreview:
    _value: 'ObjectPreview'
    _key: Optional['ObjectPreview']

    @property
    def key(self) -> 'ObjectPreview':
        return self._key

    @key.setter
    def key(self, data: dict) -> None:
        self._key = ObjectPreview(**data)

    @property
    def value(self) -> 'ObjectPreview':
        return self._value

    @value.setter
    def value(self, data: dict) -> None:
        self._value = ObjectPreview(**data)


@dataclass
class CustomPreview:
    header: str
    bodyGetterId: Optional[str] = None


@dataclass
class ObjectPreview:
    type: Literal["object", "function", "undefined", "string", "number", "boolean", "symbol", "bigint"]
    overflow: bool
    properties: List['PropertyPreview']
    entries: List['EntryPreview']
    subtype: Optional[Literal[
        "array", "null", "node", "regexp", "date", "map", "set", "weakmap", "weakset", "iterator", "generator",
        "error", "proxy", "promise", "typedarray", "arraybuffer", "dataview", "webassemblymemory", "wasmvalue"]] = None
    description: Optional[str] = None
    _entries: List['EntryPreview'] = field(init=False, repr=False, default=None)

    @property
    def properties(self) -> List['PropertyPreview']:
        return self._properties

    @properties.setter
    def properties(self, data: List[dict]) -> None:
        self._properties = [PropertyPreview(**item) for item in data]

    @property
    def entries(self) -> List['EntryPreview']:
        return self._entries

    @entries.setter
    def entries(self, data: List[dict]) -> None:
        self._entries = [EntryPreview(**item) for item in data] if not isinstance(data, property) else None


@dataclass
class RemoteObject:
    """
    Зеркальный объект, ссылающийся на исходный объект JavaScript.
    # https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#type-RemoteObject
    """
    type: Literal["object", "function", "undefined", "string", "number", "boolean", "symbol", "bigint"]
    preview: Optional["ObjectPreview"]
    customPreview: Optional["CustomPreview"]
    subtype: Optional[Literal[
        "array", "null", "node", "regexp", "date", "map", "set", "weakmap", "weakset", "iterator", "generator",
        "error", "proxy", "promise", "typedarray", "arraybuffer", "dataview", "webassemblymemory", "wasmvalue"]] = None
    className: Optional[str] = None
    value: Optional[any] = None
    unserializableValue: Optional[str] = None   # ? Примитивное значение, которое не может быть преобразовано в строку
                                                # ?     JSON, не имеет value, но получает это свойство.
    description: Optional[str] = None
    objectId: Optional[str] = None
    _preview: Optional['ObjectPreview'] = field(init=False, repr=False, default=None)
    _customPreview: Optional['CustomPreview'] = field(init=False, repr=False, default=None)

    @property
    def preview(self) -> 'ObjectPreview':
        return self._preview

    @preview.setter
    def preview(self, data: dict) -> None:
        self._vpreview = ObjectPreview(**data) if not isinstance(data, property) else None

    @property
    def customPreview(self) -> 'CustomPreview':
        return self._customPreview

    @customPreview.setter
    def customPreview(self, data: dict) -> None:
        self._customPreview = CustomPreview(**data) if not isinstance(data, property) else None


@dataclass
class AuxData:
    isDefault: bool
    type: str
    frameId: str


@dataclass
class ContextDescription:
    id: int
    origin: str     # url
    name: str
    uniqueId: str
    auxData: 'AuxData'

    @property
    def auxData(self) -> 'AuxData':
        return self._auxData

    @auxData.setter
    def auxData(self, data: dict) -> None:
        self._auxData = AuxData(**data)

@dataclass
class PropertyDescriptor:
    name: str
    configurable: bool
    enumerable: bool
    value: Optional['RemoteObject'] = None
    writable: Optional[bool] = None
    get: Optional['RemoteObject'] = None
    set: Optional['RemoteObject'] = None
    wasThrown: Optional[bool] = None
    isOwn: Optional[bool] = None
    symbol: Optional['RemoteObject'] = None



class Script:
    """
    Упаковывает вызываемое в контексте указанного фрейма выражение. Если контекст не указан, в его качестве
        будет выбран фрейм верхнего уровня. Выражение можно заменить при вызове.
    """
    def __init__(
            self, conn, expression: str,
            context: Optional[Union['ContextDescription', str]] = None):
        self._connection = conn
        self.expression = expression
        self.unique_context_id = context if type(context) is str else context.uniqueId if context is not None else None

    async def Call(
            self, expression: Optional[str] = None, returnByValue: Optional[bool] = None) -> 'RemoteObject':
        if expression:
            self.expression = expression
        args = {"expression": self.expression}
        if returnByValue is not None:
            args.update(returnByValue=returnByValue)
        if self.unique_context_id is not None:
            args.update(uniqueContextId=self.unique_context_id)
        result: dict = await self._connection.call("Runtime.evaluate", args)
        return RemoteObject(**result.get("result"))


class ContextManager:
    contexts: List['ContextDescription'] = []
    is_watch: bool = False

    async def on_create(self, data: dict) -> None:
        self.contexts.append(ContextDescription(**data.get("context")))

    async def on_clear(self, data: dict) -> None:
        self.contexts = []

    async def on_destroy(self, data: dict) -> None:
        context_id: int = data.get("executionContextId")
        ii = -1
        for i, ctx in enumerate(self.contexts):
            if ctx.id == context_id:
                ii = i
                break

        if ii > -1: self.contexts.pop(ii)

    def GetDefaultContext(self, frameId: str) -> Union['ContextDescription', None]:
        for ctx in self.contexts:
            print("frameId", frameId)
            print("Runtime context", ctx, len(self.contexts))
            if ctx.auxData.frameId == frameId and ctx.auxData.isDefault:
                return ctx
        return None



from ...data import DomainEvent


class Log:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Log
    #   LogEntry -> https://chromedevtools.github.io/devtools-protocol/tot/Log#type-LogEntry
    """
    __slots__ = ("_connection", "enabled")

    def __init__(self, conn) -> None:
        from ...connection import Connection

        self._connection: Connection = conn
        self.enabled = False

    async def enable(self) -> None:
        """
        Включает 'Log' домен, отправляет записи лога, собранные на данный момент, посредством
            события 'entryAdded'.
        https://chromedevtools.github.io/devtools-protocol/tot/Log#method-enable
        :return:
        """
        if not self.enabled:
            await self._connection.call("Log.enable")
            self.enabled = True

    async def disable(self) -> None:
        """
        Выключает 'Log' домен, останавливая отправку сообщений.
        https://chromedevtools.github.io/devtools-protocol/tot/Log#method-disable
        :return:
        """
        if self.enabled:
            await self._connection.call("Log.disable")
            self.enabled = False

    async def clear(self) -> None:
        """
        Очищает список ранее опубликованных сообщений лога.
        https://chromedevtools.github.io/devtools-protocol/tot/Log#method-clear
        :return:
        """
        await self._connection.call("Log.clear")

class LogEvent(DomainEvent):
    entryAdded = "Log.entryAdded"



import time
from .types import TouchPoint
from typing import Optional, List, Literal


class Input:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Input
    """
    __slots__ = ("_connection",)

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn

# region [ |>*<|=== Domains ===|>*<| ] Input [ |>*<|=== Domains ===|>*<| ]
    #   https://godoc.org/github.com/unixpickle/muniverse/chrome

    async def dispatchKeyEvent(
        self, type_: str,
        modifiers:             int = 0,
        timestamp:             Optional[int] = None,
        text:                  str = "",
        unmodifiedText:        str = "",
        keyIdentifier:         str = "",
        code:                  str = "",
        key:                   str = "",
        windowsVirtualKeyCode: int = 0,
        nativeVirtualKeyCode:  int = 0,
        autoRepeat:           bool = False,
        isKeypad:             bool = True,
        isSystemKey:          bool = False,
        location:              int = 0,
        commands:             Optional[list] = None
    ) -> None:
        """
        Отправляет событие кнопки на страницу.
        https://chromedevtools.github.io/devtools-protocol/tot/Input#method-dispatchKeyEvent
        :param type_:                   Тип события. Допустимые значения:
                                            keyDown, keyUp, rawKeyDown, char
        :param modifiers:               (optional) Битовое поле, представляющее нажатые клавиши-модификаторы.
                                            Alt = 1, Ctrl = 2, Meta/Command = 4, Shift = 8 (по умолчанию: 0).
        :param timestamp:               (optional) UNIX-time (число секунд с 1 января 1970)
        :param text:                    (optional) Текст генерируется путем обработки кода виртуальной клавиши
                                            с раскладкой клавиатуры. Не требуется для событий keyUp и rawKeyDown
                                            (по умолчанию: "").
        :param unmodifiedText:          (optional) Текст, который был бы сгенерирован клавиатурой, если бы не
                                            были нажаты модификаторы (кроме Shift). Полезно для обработки
                                            сочетаний клавиш (по умолчанию: "").
        :param keyIdentifier:           (optional) Уникальный идентификатор кнопки (например, 'U+0041')
                                            (по умолчанию: "").
        :param code:                    (optional) Уникальное значение строки, определяемое DOM для каждого
                                            физического ключа (например, 'KeyA') (по умолчанию: "").
        :param key:                     (optional) Уникальное строковое значение, определяемое DOM, описывающее
                                            значение клавиши в контексте активных модификаторов, раскладки
                                            клавиатуры и т. д. (Например, 'AltGr') (по умолчанию: "").
        :param windowsVirtualKeyCode:   (optional) Код виртуальной клавиши Windows (по умолчанию: 0).
        :param nativeVirtualKeyCode:    (optional) Собственный код виртуальной клавиши (по умолчанию: 0).
        :param autoRepeat:              (optional) Было ли событие сгенерировано из автоматического повтора
                                            (по умолчанию: false).
        :param isKeypad:                (optional) Было ли событие сгенерировано с клавиатуры (по умолчанию: false).
        :param isSystemKey:             (optional) Было ли событие, событием системной кнопки (по умолчанию: false).
        :param location:                (optional) Было ли событие с левой или правой стороны клавиатуры.
                                            1 = слева, 2 = справа (по умолчанию: 0).
        :param commands:                (optional) Команды редактирования для отправки с ключевым событием (например,
                                            'selectAll') (по умолчанию: []). Они связаны с именами команд,
                                            используемыми в document.execCommand и NSStandardKeyBindingResponding,
                                            но не равны им.(по умолчанию: []).
        :return:        None
        """
        args = {
            "type": type_, "modifiers": modifiers, "text": text, "unmodifiedText": unmodifiedText,
            "keyIdentifier": keyIdentifier, "code": code, "key": key,
            "windowsVirtualKeyCode": windowsVirtualKeyCode, "nativeVirtualKeyCode": nativeVirtualKeyCode,
            "autoRepeat": autoRepeat, "isKeypad": isKeypad, "isSystemKey": isSystemKey, "location": location,
            "commands": (commands if commands else [])
        }
        if timestamp is not None:
            args.update({"timestamp": timestamp})
        else:
            args.update({"timestamp": int(time.time() * 1000)})
        await self._connection.call("Input.dispatchKeyEvent", args)

    async def dispatchMouseEvent(
        self, type_: str, x: float, y: float,
        modifiers:            int = 0,
        timestamp:            Optional[int] = None,
        button:               str = "none",
        buttons:              int = 0,
        clickCount:           int = 1,
        force:              float = 0,
        tangentialPressure: float = 0,
        tiltX:                int = 0,
        tiltY:                int = 0,
        twist:                int = 0,
        deltaX:             float = 0,
        deltaY:             float = 0,
        pointerType:          str = "mouse"
    ) -> None:
        """
        Отправляет событие мышки на страницу.
        https://chromedevtools.github.io/devtools-protocol/tot/Input#method-dispatchMouseEvent
        :param type_:                   Тип события. Допустимые значения:
                                            mousePressed, mouseReleased, mouseMoved, mouseWheel
        :param x:                       Координата X события относительно области просмотра основного
                                            фрейма в пикселях CSS.
        :param y:                       Координата Y события относительно области просмотра основного
                                            фрейма в пикселях CSS.
        :param modifiers:               (optional) Битовое поле, представляющее нажатые клавиши-модификаторы.
                                            Alt = 1, Ctrl = 2, Meta/Command = 4, Shift = 8 (по умолчанию: 0).
        :param timestamp:               (optional) UNIX-time (число секунд с 1 января 1970)
        :param button:                  (optional) Кнопка мыши. Допустимые значения:
                                            none, left, middle, right, back, forward
        :param buttons:                 (optional) Число, указывающее, какие кнопки нажимаются на мыши, когда
                                            вызывается событие мыши.
                                            Left=1, Right=2, Middle=4, Back=8, Forward=16, None=0.
        :param clickCount:              (optional) Количество нажатий кнопки мыши (по умолчанию: 0).

        ========================== опции стилуса ========================
        :param force:                   (optional, EXPERIMENTAL) - Нормализованное давление(на кнопку?) в
                                            интервале от 0 до 1.
        :param tangentialPressure:      (optional, EXPERIMENTAL) - Нормализованное тангенциальное давление
                                            (на кнопку?) в интервале от -1 до 1.
        :param tiltX:                   (optional, EXPERIMENTAL) - Плоский угол между плоскостью YZ и плоскостью,
                                            содержащей ось пера и ось Y, в градусах диапазона [-90,90], положительный
                                            tiltX направлен вправо (по умолчанию: 0).
        :param tiltY:                   (optional, EXPERIMENTAL) - Плоский угол между плоскостью XZ и плоскостью,
                                            содержащей ось пера и ось X, в градусах диапазона [-90,90], положительный
                                            tiltY направлен к пользователю (по умолчанию: 0).
        :param twist:                   (optional, EXPERIMENTAL) - Вращение стилуса пера по часовой стрелке вокруг
                                            своей главной оси в градусах в диапазоне [0,359] (по умолчанию: 0).
        ========================== конец опций стилуса ========================

        :param deltaX:                  (optional) Дельта X в пикселях CSS для события колеса мыши
                                            (по умолчанию: 0).
        :param deltaY:                  (optional) Дельта Y в пикселях CSS для события колеса мыши
                                            (по умолчанию: 0).
        :param pointerType:             (optional) Тип указателя (по умолчанию: "mouse"). Допустимые значения:
                                            mouse, pen
        :return:
        """
        args = {
            "type": type_, "x": x, "y": y, "modifiers": modifiers, "button": button, "buttons": buttons,
            "clickCount": clickCount, "force": force, "tangentialPressure": tangentialPressure, "tiltX": tiltX,
            "tiltY": tiltY, "twist": twist, "deltaX": deltaX, "deltaY": deltaY, "pointerType": pointerType
        }
        if timestamp is not None:
            args.update({"timestamp": timestamp})
        else:
            args.update({"timestamp": int(time.time() * 1000)})
        await self._connection.call("Input.dispatchMouseEvent", args)

    async def dispatchTouchEvent(
        self, type_: str,
        touchPoints: Optional[List[TouchPoint]] = None,
        modifiers:                int = 0,
        timestamp:                Optional[int] = None
    ) -> None:
        """
        Отправляет событие прикосновения на страницу.
        https://chromedevtools.github.io/devtools-protocol/tot/Input#method-dispatchTouchEvent
        :param type_:                   Тип сенсорного события. TouchEnd и TouchCancel не должны содержать
                                            никаких точек касания, в то время как TouchStart и TouchMove
                                            должны содержать хотя бы одну. Допустимые значения:
                                            touchStart, touchEnd, touchMove, touchCancel
        :param touchPoints:             Список точек касания на сенсорном устройстве. Генерируется одно
                                            событие на каждую измененную точку (по сравнению с предыдущим
                                            событием касания в последовательности), эмулируя
                                            нажатия/перемещения/отпускания точек одну за другой.
        :param modifiers:               (optional) Битовое поле, представляющее нажатые клавиши-модификаторы.
                                            Alt = 1, Ctrl = 2, Meta/Command = 4, Shift = 8 (по умолчанию: 0).
        :param timestamp:               (optional) UNIX-time (число секунд с 1 января 1970)
        :return:
        """
        args = { "type": type_, "modifiers": modifiers }
        if type_ == "touchEnd" or type_ == "touchCancel":
            if len(touchPoints) == 0 or touchPoints is None:
                raise ValueError(f"Для действия '{type_}', аргумент 'touchPoints' должен получить хотя-бы одну координату")
            args.update({"touchPoints": [point.to_dict() for point in touchPoints]})

        if timestamp is not None: args.update({"timestamp": timestamp})
        else: args.update({"timestamp": int(time.time() * 1000)})

        await self._connection.call("Input.dispatchTouchEvent", args)

    async def insertText(self, text: str) -> None:
        """
        (EXPERIMENTAL)
        Этот метод эмулирует вставку текста, который не поступает от нажатия клавиши, например,
            клавиатуры Emoji или IME.
        https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-insertText
        :param text:                    Текст для вставки.
        :return:
        """
        await self._connection.call("Input.insertText", {"text": text})

    async def synthesizePinchGesture(
            self, x: float, y: float, scaleFactor: float,
                relativeSpeed: Optional[int] = None,
            gestureSourceType: Optional[Literal["touch", "mouse"]] = None
    ) -> None:
        """
        (EXPERIMENTAL)
        Синтезирует жест щипка за период времени, генерируя соответствующие сенсорные события.
        https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-synthesizePinchGesture
        :param x:                       Координата X начала жеста в пикселях CSS.
        :param y:                       Координата Y начала жеста в пикселях CSS.
        :param scaleFactor:             Относительный масштабный коэффициент после увеличения
                                            (> 1,0 увеличивает, <1,0 уменьшает).
        :param relativeSpeed:           (optional) Относительная скорость указателя в пикселях в секунду (по умолчанию: 800).
        :param gestureSourceType:       (optional) Тип входных событий, который должен быть сгенерирован
                                            (по умолчанию: "default", который запрашивает у платформы
                                            предпочтительный тип ввода). default, touch, mouse
        :return:
        """
        args = {"x": x, "y": y, "scaleFactor": scaleFactor}
        if relativeSpeed is not None:
            args.update({"relativeSpeed": relativeSpeed})
        if gestureSourceType is not None:
            args.update({"gestureSourceType": gestureSourceType})
        await self._connection.call("Input.synthesizePinchGesture", args)

    async def synthesizeScrollGesture(
            self, x: float, y: float,
                      xDistance: Optional[float] = None,
                      yDistance: Optional[float] = None,
                    xOverscroll: Optional[float] = None,
                    yOverscroll: Optional[float] = None,
                    preventFling: Optional[bool] = None,
                            speed: Optional[int] = None,
                gestureSourceType: Optional[Literal["touch", "mouse"]] = None,
                      repeatCount: Optional[int] = None,
                    repeatDelayMs: Optional[int] = None,
            interactionMarkerName: Optional[str] = None
    ) -> None:
        """
        (EXPERIMENTAL)
        Синтезирует жест прокрутки в течение определенного периода времени, генерируя соответствующие сенсорные события.
            Возвращает управление только после выполнения жеста!
        https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-synthesizeScrollGesture
        :param x:                       Координата X начала жеста в пикселях CSS.
        :param y:                       Координата Y начала жеста в пикселях CSS.
        :param xDistance:               (optional) Расстояние для прокрутки вдоль оси X (положительное для прокрутки влево).
        :param yDistance:               (optional) Расстояние для прокрутки вдоль оси Y (положительное для прокрутки вверх).
        :param xOverscroll:             (optional) Количество дополнительных пикселей для прокрутки назад вдоль оси X, в
                                            дополнение к заданному расстоянию.
        :param yOverscroll:             (optional) Количество дополнительных пикселей для прокрутки назад вдоль оси Y, в
                                            дополнение к заданному расстоянию.
        :param preventFling:            (optional) Запретить сброс (по умолчанию: true).
        :param speed:                   (optional) Скорость прокрутки в пикселях в секунду (по умолчанию: 800).
        :param gestureSourceType:       (optional) Тип входных событий, который должен быть сгенерирован
                                            (по умолчанию: "default", который запрашивает у платформы
                                            предпочтительный тип ввода). default, touch, mouse
        :param repeatCount:             (optional) Количество повторений жеста (по умолчанию: 0).
        :param repeatDelayMs:           (optional) Задержка в миллисекундах между повторениями. (по умолчанию: 250).
        :param interactionMarkerName:   (optional) Имя генерируемых маркеров взаимодействия, если оно не пустое (по умолчанию: "").
        :return:
        """
        args = {"x": x, "y": y}
        if xDistance is not None:
            args.update({"xDistance": xDistance})
        if yDistance is not None:
            args.update({"yDistance": yDistance})
        if xOverscroll is not None:
            args.update({"xOverscroll": xOverscroll})
        if yOverscroll is not None:
            args.update({"yOverscroll": yOverscroll})
        if preventFling is not None:
            args.update({"preventFling": preventFling})
        if speed is not None:
            args.update({"speed": speed})
        if gestureSourceType is not None:
            args.update({"gestureSourceType": gestureSourceType})
        if repeatCount is not None:
            args.update({"repeatCount": repeatCount})
        if repeatDelayMs is not None:
            args.update({"repeatDelayMs": repeatDelayMs})
        if interactionMarkerName is not None:
            args.update({"interactionMarkerName": interactionMarkerName})
        await self._connection.call("Input.synthesizeScrollGesture", args)

    async def synthesizeTapGesture(
            self, x: float, y: float,
            duration:          Optional[int] = None,
            tapCount:          Optional[int] = None,
            gestureSourceType: Optional[Literal["touch", "mouse"]] = None
    ) -> None:
        """
        (EXPERIMENTAL)
        Синтезирует жест касания за период времени, генерируя соответствующие сенсорные события.
        https://chromedevtools.github.io/devtools-protocol/tot/Input/#method-synthesizeTapGesture
        :param x:                       Координата X места касания в пикселях CSS.
        :param y:                       Координата Y места касания в пикселях CSS.
        :param duration:                (optional) Продолжительность касания (по умолчанию: 50).
        :param tapCount:                (optional) Количество прикосновений (по умолчанию: 1).
        :param gestureSourceType:       (optional) Тип входных событий, который должен быть сгенерирован
                                            (по умолчанию: "default", который запрашивает у платформы
                                            предпочтительный тип ввода). default, touch, mouse
        :return:
        """
        args = {"x": x, "y": y}
        if duration is not None:
            args.update({"duration": duration})
        if tapCount is not None:
            args.update({"tapCount": tapCount})
        if gestureSourceType is not None:
            args.update({"gestureSourceType": gestureSourceType})
        await self._connection.call("Input.synthesizeTapGesture", args)

    # endregion



from dataclasses import dataclass
from typing import Optional


@dataclass
class TouchPoint:
    """ Описывает точку прикосновения стилуса для метода DispatchTouchEvent() """
    x: float
    y: float                                # Координаты точки касания
    radiusX: Optional[float] = None         # X-радиус прикосновения(по умолчанию 1.0)
    radiusY: Optional[float] = None         # Y-радиус прикосновения(по умолчанию 1.0)
    rotationAngle: Optional[float] = None   # Угол вращения(по умолчанию 0.0)
    force: Optional[float] = None           # Сила(по умолчанию 1.0)
    tangentialPressure: Optional[float] = None  # Нормализованное тангенциальное давление в
                                                #   диапазоне [-1,1](по умолчанию 0.0)
    tiltX: Optional[int] = None             # Угол между плоскостью Y-Z и плоскостью,
                                                #   содержащей ось стилуса и ось Y, в градусах
                                                #   диапазона [-90,90], положительный tiltX
                                                #   - наклон вправо (по умолчанию: 0).
    tiltY: Optional[int] = None             # Угол между плоскостью X-Z и плоскостью,
                                                #   содержащей ось стилуса и ось X, в градусах
                                                #   диапазона [-90,90], положительный tiltY
                                                #   - наклон вниз (по умолчанию: 0).
    twist: Optional[int] = None             # Поворот стилуса пера по часовой стрелке вокруг
                                                #   своей главной оси в градусах в диапазоне
                                                #   [0,359] (по умолчанию: 0).
    id: Optional[float] = None              # Идентификатор, используемый для отслеживания
                                                #   источников касания между событиями, должен
                                                #   быть уникальным в пределах события.
    def to_dict(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if v is not None}



try:
    import ujson as json
except ModuleNotFoundError:
    import json

from .actions import Actions
from .data import ViewportRect, WindowRect, GeoInfo

import base64, re
from typing import Optional

from .exceptions import EvaluateError, JavaScriptError, NullProperty


class Extend:
    """ Расширение для 'Page' некоторыми полезными методами.
    """
    __slots__ = ("_connection", "action", "_py_call_script_id")

    def __init__(self, conn) -> None:

        from .connection import Connection

        self._connection: Connection = conn
        self._py_call_script_id: str = ""
        self.action = Actions(conn)             # Совершает действия на странице. Клики;
                                                # движения мыши; события клавиш

    @property
    def py_call_enabled(self) -> bool:
        """ Был ли обработчик `py_call()` зарегистрирован на страницу.
        """
        return bool(self._py_call_script_id)

    async def pyCallAddOnload(self) -> None:
        """ Включает автоматически добавляющийся JavaScript, вызывающий слушателей
        клиента, добавленных на страницу с помощью await <Connection>.bindFunction(...)
        и await <Connection>.bindFunctions(...).

        Например, `test_func()` объявленная следующим образом:
        async def test_func(number: int, text: str, bind_arg: dict) -> None:
            print("[- test_func -] Called with args:\n\tnumber: "
                  f"{number}\n\ttext: {text}\n\tbind_arg: {bind_arg}")

        И добавленная  следующим образом:
        await conn.bindFunction(
            test_func,                          # ! слушатель
            {"name": "test", "value": True}     # ! bind_arg
        )

        Может быть вызвана со страницы браузера, так:
        py_call("test_func", 1, "testtt");
        """
        if self.py_call_enabled:
            return

        py_call_js = """\
        function py_call(funcName,...args){eval(funcName+"(`"+JSON.stringify(args)+"`)");}"""
        self._py_call_script_id = await self._connection.Page.addScriptOnLoad(py_call_js)
        await self.injectJS(py_call_js)

    async def pyCallRemoveOnLoad(self) -> None:
        """ Удаляет автоматическое добавление JavaScript, установленного
        pyCallAddOnload().
        """
        if self.py_call_enabled:
            await self._connection.Page.removeScriptOnLoad(self._py_call_script_id)
            self._py_call_script_id = ""


    async def getViewportRect(self) -> ViewportRect:
        """ Возвращает список с длиной и шириной вьюпорта браузера.
        """
        code = "(()=>{return JSON.stringify([window.innerWidth,window.innerHeight]);})();"
        data = json.loads(await self.injectJS(code))
        return ViewportRect(int(data[0]), int(data[1]))

    async def getWindowRect(self) -> WindowRect:
        """ Возвращает список с длиной и шириной окна браузера.
        """
        code = "(()=>{return JSON.stringify([window.outerWidth,window.outerHeight]);})();"
        data = json.loads(await self.injectJS(code))
        return WindowRect(int(data[0]), int(data[1]))

    async def getUrl(self) -> str:
        return (await self._connection.Target.getTargetInfo()).url

    async def getTitle(self) -> str:
        return (await self._connection.Target.getTargetInfo()).title

    async def makeScreenshot(
            self,
                 format_: str = "",
                 quality: int = -1,
                   clip: Optional[dict] = None,
            fromSurface: bool = True
    ) -> bytes:
        """  Сделать скриншот. Возвращает набор байт, представляющий скриншот.
        :param format_:         jpeg или png (по умолчанию png).
        :param quality:         Качество изображения в диапазоне [0..100] (только для jpeg).
        :param clip:            {
                                    "x": "number => X offset in device independent pixels (dip).",
                                    "y": "number => Y offset in device independent pixels (dip).",
                                    "width": "number => Rectangle width in device independent pixels (dip).",
                                    "height": "number => Rectangle height in device independent pixels (dip).",
                                    "scale": "number => Page scale factor."
                                }
        :param fromSurface:     boolean => Capture the screenshot from the surface, rather than the view.
                                    Defaults to true.
        :return:                bytes
        """
        shot = await self._connection.Page.captureScreenshot(format_, quality, clip, fromSurface)
        return base64.b64decode(shot.encode("utf-8"))

    async def selectOption(self, css: str) -> None:
        """ Создаёт фокус и делает выбранным опцию тега <select>
        при помощи JavaScript.
        """
        await self.injectJS(f"let _i_ = document.querySelector(`{css}`); _i_.focus(); _i_.select();")

    async def scrollIntoViewJS(self, selector: str) -> None:
        """ Выполняет плавную прокрутку страницы до выбранного селектора. """
        await self.injectJS(
            "document.querySelector(`" +
            selector +
            "`).scrollIntoView({'behavior':'smooth', 'block': 'center'});"
        )

    async def evalPromise(self, expression: str) -> dict:
        """ Выполняет асинхронный код на странице и дожидается
        получения результата.
        """
        result = await self._connection.Runtime.evaluate(
            expression=expression,
            objectGroup="console",
            includeCommandLineAPI=True,
            silent=False,
            returnByValue=False,
            userGesture=True,
            awaitPromise=False
        )

        response = await self._connection.Runtime.awaitPromise(
            promiseObjectId=result.objectId,
            returnByValue=False,
            generatePreview=False
        )

        return json.loads(response.value)

    async def injectJS(self, expression: str) -> any:
        """ Выполняет JavaScript-выражение во фрейме верхнего уровня. """
        try:
            response = await self._connection.Runtime.evaluate(
                expression=expression,
                objectGroup="console",
                includeCommandLineAPI=True,
                silent=False,
                returnByValue=False,
                userGesture=True,
                awaitPromise=False
            )
        except EvaluateError as error:
            error = str(error)
            if "of null" in error:
                if match := re.match(r"[\w\s:]+['|\"]([^'\"]+)", error):
                    prop = match.group(1)
                else:
                    prop = "unmatched error: " + error
                raise NullProperty("InjectJS() Exception with injected code:"
                                   f"\n'{expression}'\nNull property:\n{prop}")

            raise JavaScriptError("JavaScriptError: InjectJS() Exception with "
                                  f"injected code:\n'{expression}'\nDescription:\n{error}")

        return response.value

    async def getGeoInfo(self) -> GeoInfo:
        """ Возвращает информацию о местоположении точки выхода браузера в сеть,
        вычисленному по IP.
        """
        async_fn_js = """\
        async function get_geo_info() {
            const resp = await fetch('https://time.gologin.com/');
            return await resp.text();
        } get_geo_info();
        """

        promise = """fetch('https://time.gologin.com/').then(res => res.text())"""

        result: dict = await self._connection.extend.evalPromise(promise)
        result.update(
            geo=dict(
                latitude=float(result["ll"][0]),
                longitude=float(result["ll"][1]),
                accuracy=float(result["accuracy"])
            ),
            languages=result["languages"].split(","),
            state_province=result.get("stateProv"),
            proxy_type=(pt:=result.get("proxyType"))
        )
        del result["ll"]
        del result["accuracy"]
        del result["stateProv"]
        if pt is not None:
            del result["proxyType"]
        return GeoInfo(**result)



try:
    import ujson as json
except ModuleNotFoundError:
    import json
import asyncio
from websockets.client import WebSocketClientProtocol, connect
from websockets.exceptions import ConnectionClosedError
from inspect import iscoroutinefunction
from typing import Callable, Optional, Union, Tuple, Dict, Any, Iterable, Coroutine

from .exceptions import get_cdtp_error
from .utils import log

from .data import DomainEvent, Sender, Receiver, CommonCallback
from .extend_connection import Extend

from .domains.background_service import BackgroundService
from .domains.browser import Browser
from .domains.css import CSS
from .domains.device_orientation import DeviceOrientation
from .domains.dom import DOM
from .domains.emulation import Emulation
from .domains.fetch import Fetch
from .domains.input import Input
from .domains.log import Log
from .domains.network import Network
from .domains.overlay import Overlay
from .domains.page import Page
from .domains.runtime import Runtime
from .domains.system_info import SystemInfo
from .domains.target import Target

CoroTypeNone = Coroutine[None, None, None]


class Connection:
    """ Если инстанс страницы более не нужен, например, при перезаписи в него нового
    инстанса, перед этим [-!-] ОБЯЗАТЕЛЬНО [-!-] - вызовите у него метод
    Detach(), или закройте вкладку/страницу браузера, с которой он связан,
    тогда это будет выполнено автоматом. Иначе в цикле событий останутся
    задачи связанные с поддержанием соединения, которое более не востребовано.
    """
    __slots__ = (
        "ws_url", "frontend_url", "callback", "_id", "extend", "_bindings",
        "responses", "_ws_session", "_receiver_loop", "_on_detach_listener", "_listeners_for_event",
        "on_close_event", "context_manager", "_connected", "_conn_id", "_verbose",
        "_browser_name", "_is_headless_mode",

        "BackgroundService", "Browser", "CSS", "DeviceOrientation", "DOM", "Emulation", "Fetch", "Input",
        "Log", "Network", "Overlay", "Page", "Runtime", "SystemInfo", "Target",
    )

    def __init__(
            self,
            ws_url: str,
            conn_id: str,
            frontend_url: str,
            callback: CommonCallback,
            is_headless_mode: bool,
            verbose: bool,
            browser_name: str
    ) -> None:
        """
        :param ws_url:              Адрес WebSocket
        :param conn_id:             Идентификатор страницы
        :param frontend_url:        devtoolsFrontendUrl по которому происходит подключение к дебаггеру
        :param callback:            Колбэк, который будет получать все данные,
                                        приходящие по WebSocket в виде словарей
        :param is_headless_mode:    "Headless" включён?
        :param verbose:             Печатать некие подробности процесса?
        :param browser_name:        Имя браузера
        """

        self.ws_url = ws_url
        self.frontend_url = frontend_url
        self.callback = callback
        self._is_headless_mode = is_headless_mode
        self._conn_id = conn_id
        self._verbose = verbose
        self._browser_name = browser_name
        self._id = 0
        self._connected = False
        self._ws_session: Optional[WebSocketClientProtocol] = None
        self._receiver_loop: Optional[asyncio.Task] = None
        self._on_detach_listener: Tuple[Callable[[any], CoroTypeNone], list, dict] = tuple()
        self._bindings: Dict[str, Tuple[Callable[[any], CoroTypeNone], Tuple[Any, ...]]] = {}
        self._listeners_for_event: Dict[
            str, Dict[
                Callable[[dict, tuple], CoroTypeNone],
                Tuple[Any, ...]
            ]
        ] = {}
        self.on_close_event = asyncio.Event()
        self.responses: Dict[int, Optional[Sender[dict]]] = {}

        self.extend = Extend(self)

        self.BackgroundService = BackgroundService(self)
        self.Browser = Browser(self)
        self.CSS = CSS(self)
        self.DeviceOrientation = DeviceOrientation(self)
        self.DOM = DOM(self)
        self.Emulation = Emulation(self)
        self.Fetch = Fetch(self)
        self.Input = Input(self)
        self.Log = Log(self)
        self.Network = Network(self)
        self.Overlay = Overlay(self)
        self.Page = Page(self)
        self.Runtime = Runtime(self)
        self.SystemInfo = SystemInfo(self)
        self.Target = Target(self)


    @property
    def connected(self) -> bool:
        return self._connected

    @property
    def conn_id(self) -> str:
        return self._conn_id

    @property
    def verbose(self) -> bool:
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool) -> None:
        self._verbose = value

    @property
    def browser_name(self) -> str:
        return self._browser_name

    @property
    def is_headless_mode(self) -> bool:
        return self._is_headless_mode

    def __str__(self) -> str:
        return f"<Connection targetId={self.conn_id!r}>"

    def __eq__(self, other: "Connection") -> bool:
        return self.conn_id == other.conn_id

    def __hash__(self) -> int:
        return hash(self.conn_id)

    async def call(
        self, domain_and_method: str,
        params:  Optional[dict] = None,
        wait_for_response: bool = True
    ) -> Optional[dict]:
        self._id += 1
        _id = self._id
        data = {
            "id": _id,
            "params": params if params else {},
            "method": domain_and_method
        }

        if not wait_for_response:
            self.responses[_id] = None
            await self._send(json.dumps(data))
            return

        que = asyncio.Queue()
        sender, receiver = Sender[dict](que), Receiver[dict](que)
        self.responses[_id] = sender

        await self._send(json.dumps(data))

        response = await receiver.recv()
        if "error" in response:

            if ex := get_cdtp_error((e := response['error'])['message']):
                raise ex(f"domain_and_method = '{domain_and_method}' | params = '{str(params)}'")

            raise Exception(
                "\x1b[36mBrowser detect error:\n" +
                f"\x1b[37mError code: '\x1b[91m{e['code']}\n" +
                f"\x1b[37mError message: '\x1b[91m{e['message']}\n" +
                f"\x1b[37mdomain_and_method: '\x1b[91m{domain_and_method}\n" +
                f"\x1b[37mparams: '\x1b[91m{params}\x1b[0m\n"
            )

        return response["result"]

    async def _send(self, data: str) -> None:
        if self.connected:
            await self._ws_session.send(data)

    async def _recv(self) -> None:
        while self.connected:
            try:
                data_msg: dict = json.loads(await self._ws_session.recv())
            # ! Браузер разорвал соединение
            except ConnectionClosedError as e:
                if self.verbose: log(f"ConnectionClosedError {e!r}")
                await self.detach()
                return

            # Ожидающие ответов вызовы API получают ответ по id входящих сообщений.
            if sender := self.responses.pop(data_msg.get("id"), None):
                await sender.send(data_msg)

            if ((method := data_msg.get("method")) == "Inspector.detached"
                    and data_msg["params"]["reason"] == "target_closed"):
                await self.detach()
                self.on_close_event.set()
                return

            # Если коллбэк функция была определена, она будет получать все
            #   уведомления из инстанса страницы.
            if self.callback is not None:
                asyncio.create_task(self.callback(data_msg))

            # ? Был вызов из контекста страницы
            if method == "Runtime.bindingCalled":
                name: str = data_msg["params"]["name"]
                payload: str = data_msg["params"]["payload"]

                # ? Есть вызываемый объект с таким именем
                if handle := self._bindings.get(name):
                    function, args = handle
                    asyncio.create_task(
                        function(
                            *json.loads(payload),
                            *args
                        )
                    )

            if listeners := self._listeners_for_event.get(method):
                p = data_msg.get("params") or {}
                for listener, largs in listeners.items():
                    asyncio.create_task(
                        listener(       # корутина
                            p,          # её "params" — всегда передаётся
                            *largs      # список bind-агрументов
                        )
                    )



    async def waitForClose(self) -> None:
        """ Дожидается, пока не будет потеряно соединение со страницей. """
        await self.on_close_event.wait()

    async def activate(self) -> None:
        self._ws_session = await connect(self.ws_url, ping_interval=None)
        self._connected = True
        self._receiver_loop = asyncio.create_task(self._recv())
        await self.Runtime.enable()

    async def detach(self) -> None:
        """  Отключается от страницы. Вызывается автоматически при закрытии браузера,
        или текущей страницы. Принудительный вызов не закрывает страницу,
        а лишь разрывает с ней соединение.
        """
        if not self.connected:
            return

        self._receiver_loop.cancel()
        if self.verbose: log(f"[ DETACH ] {self.conn_id}")
        self._connected = False

        if self._on_detach_listener:
            function, args, kvargs = self._on_detach_listener
            await function(*args, **kvargs)

    def clearOnDetach(self) -> None:
        self._on_detach_listener = tuple()

    def setOnDetach(self, function: Callable[[any], CoroTypeNone], *args, **kvargs) -> bool:
        """  Регистрирует асинхронный коллбэк, который будет вызван с соответствующими аргументами
        при разрыве соединения со страницей.
        """
        if not iscoroutinefunction(function):
            raise TypeError("OnDetach-listener must be a async callable object!")
        if not self.connected:
            return False
        self._on_detach_listener = function, args, kvargs
        return True

    async def bindFunction(self, function: Callable[[any], CoroTypeNone], *args: Any) -> None:
        """ Регистрирует имя в глобальном контексте страницы. Это имя затем используется
        в вызове функции, принимающей ровно один, строковый аргумент, который передаётся
        в тело события `Runtime.bindingCalled`.
        """
        if not iscoroutinefunction(function):
            raise TypeError("Listener must be a async callable object!")

        self._bindings[function.__name__] = function, args
        await self.Runtime.addBinding(function.__name__)
        await self.extend.pyCallAddOnload()

    async def bindFunctions(
            self, *handlers_n_args: Tuple[Callable[[any], CoroTypeNone], Iterable]) -> None:
        """ Выполняет множественную регистрацию.
        :param handlers_n_args:     Список двухэлементных последовательностей,
        в которых:
            - первый элемент: awaitable-объект
            - второй элемент: последовательность аргументов любой длины, которая
                будет передана первому элементу в последнюю очередь.
        """
        for function, args in handlers_n_args:
            await self.bindFunction(function, *args)
        await self.extend.pyCallAddOnload()

    async def unbindFunctions(self, *functions: Union[Callable[[any], CoroTypeNone], str]) -> None:
        """ Прекращает генерацию событий `Runtime.bindingCalled` для указанных имён.
        :param functions:  Список функций, или их имён.
        """
        for function in functions:
            name = function if type(function) is str else function.__name__
            self._bindings.pop(name)
            await self.Runtime.removeBinding(name)

    async def addListenerForEvent(
        self, event: Union[str, DomainEvent], listener: Callable[[dict, tuple], CoroTypeNone], *args) -> None:
        """ Регистрирует слушателя, который будет вызываться при генерации определённых событий
        в браузере. Список таких событий можно посмотреть в разделе "Events" почти
        у каждого домена по адресу: https://chromedevtools.github.io/devtools-protocol/
        Например: 'DOM.attributeModified'
            !Внимание! Каждый такой слушатель должен иметь один обязательный 'data'-аргумент,
        в который будут передаваться параметры случившегося события в виде словаря(dict).

        :param event:           Имя события, для которого регистируется слушатель. Например:
                                    'DOM.attributeModified'.
        :param listener:        Колбэк-функция.
        :param args:            (optional) любое кол-во агрументов, которые будут переданы
                                    в функцию последними.
        :return:        None
        """
        if not iscoroutinefunction(listener):
            raise TypeError("Listener must be a async callable object!")

        e: str = event if type(event) is str else event.value
        if e not in self._listeners_for_event:
            self._listeners_for_event[e]: dict = {}
        self._listeners_for_event[e][listener] = args

    def removeListenerForEvent(
            self, event: Union[str, DomainEvent], listener: Callable[[dict, tuple], CoroTypeNone]) -> None:
        """  Удаляет регистрацию слушателя для указанного события.
        :param event:           Имя метода, для которого была регистрация.
        :param listener:        Колбэк-функция, которую нужно удалить.
        :return:        None
        """
        e = event if type(event) is str else event.value
        if not iscoroutinefunction(listener):
            raise TypeError("Listener must be a async callable object!")
        if m := self._listeners_for_event.get(e):
            if listener in m: m.pop(listener)


    def removeListenersForEvent(self, event: Union[str, DomainEvent]) -> None:
        """
        Удаляет регистрацию метода и слушателей вместе с ним для указанного события.
        :param event:          Имя метода, для которого была регистрация.
        :return:        None
        """
        e = event if type(event) is str else event.value
        if e in self._listeners_for_event:
            self._listeners_for_event.pop(e)

    def __del__(self) -> None:
        if self.verbose: log(f"[ DELETED ] {self.conn_id}")



from ...data import DomainEvent

class BackgroundService:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService
    """
    __slots__ = ("_connection", "observing_started", "recording_started")

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn
        self.observing_started = False
        self.recording_started = False

    async def backgroundClearEvents(self, service: str) -> None:
        """
        Удаляет все сохраненные данные для service.
        https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/#method-clearEvents
        :param service:         Фоновая служба, которая будет связана с командами / событиями. Каждая фоновая
                                    служба работает независимо, но использует один и тот же API.
                                Допустимые значения:
                                    backgroundFetch, backgroundSync, pushMessaging, notifications,
                                    paymentHandler,periodicBackgroundSync
        :return:
        """
        await self._connection.call("BackgroundService.clearEvents", {"service": service})

    async def backgroundSetRecording(self, shouldRecord: bool, service: str) -> None:
        """
        Включает, или выключает запись для service.
        https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/#method-setRecording
        :param shouldRecord:    True == включить
        :param service:         ---------------------
        :return:
        """
        await self._connection.call("BackgroundService.clearEvents", {"shouldRecord": shouldRecord, "service": service})
        self.recording_started = shouldRecord

    async def backgroundStartObserving(self, service: str) -> None:
        """
        Включает обновления событий для service.
        https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/#method-startObserving
        :param service:         ---------------------
        :return:
        """
        if not self.observing_started:
            await self._connection.call("BackgroundService.startObserving", {"service": service})
            self.observing_started = True

    async def backgroundStopObserving(self, service: str) -> None:
        """
        Выключает обновления событий для service.
        https://chromedevtools.github.io/devtools-protocol/tot/BackgroundService/#method-stopObserving
        :param service:         ---------------------
        :return:
        """
        if self.observing_started:
            await self._connection.call("BackgroundService.stopObserving", {"service": service})
            self.observing_started = False

class BackgroundServiceEvent(DomainEvent):
    backgroundServiceEventReceived = "BackgroundService.backgroundServiceEventReceived"
    recordingStateChanged = "BackgroundService.recordingStateChanged"



from dataclasses import dataclass
from typing import Optional


@dataclass
class WindowInfo:
    windowId: int
    bounds: "Bounds"

@dataclass
class Bounds:
    """ Описывает положение, размер и состояние окна браузера """
    left: Optional[int] = None              # Смещение от левого края экрана до окна в пикселях.
    top: Optional[int] = None               # Смещение от верхнего края экрана к окну в пикселях.
    width: Optional[int] = None             # Ширина окна в пикселях.
    height: Optional[int] = None            # Высота окна в пикселях.
    windowState: Optional[str] = "normal"   # normal, minimized, maximized, fullscreen
                                        #   minimized, maximized и fullscreen нельзя сочетать с  left, top,
                                        #   width или height. Неопределенные поля останутся без изменений.

    def to_dict(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if v is not None}

@dataclass
class Version:
    protocolVersion: str
    product: str
    revision: str
    userAgent: str
    jsVersion: str



import asyncio
import subprocess
import re
import sys
import urllib.request
from typing import Optional, Dict, Callable


def get_request(url: str) -> str:
    with urllib.request.urlopen(url) as response:
        return response.read().decode('utf-8')


def find_browser_executable_path(exe="chrome") -> str:
    """ Возвращает путь до EXE. """
    if not sys.platform == "win32":
        raise OSError("registry_read_key() is only available on Windows")

    if exe == "chromium":
        import os
        from pathlib import Path

        app_data = Path(os.getenv('APPDATA')).parent
        browser_path = app_data / "Local/Chromium/Application/chrome.exe"
        if not browser_path.exists():
            return ""
        return browser_path.as_posix()

    import winreg

    reg_path = f"HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\{exe}.exe"
    key, path = re.findall(r"(^[^\\/]+)[\\/](.*)", reg_path)[0]
    connect_to = eval(f"winreg.{key}")
    try: h_key = winreg.OpenKey( winreg.ConnectRegistry(None, connect_to), path )
    except FileNotFoundError: return ""
    result = winreg.QueryValue(h_key, None)
    winreg.CloseKey(h_key)
    return result


def log(data: any = "", lvl: str = "[<- V ->]", eol: str = "\n") -> None:
    print(f"\x1b[32m{lvl} \x1b[38m\x1b[3m{data}\x1b[0m", end=eol)


def save_img_as(path: str, data: bytes) -> None:
    """ Сохраняет по пути 'path' набор байт 'data', которые можно прислать
    из метода страницы MakeScreenshot()
    """
    with open(path, "wb") as f:
        f.write(data)


async def async_util_call(function: Callable, *args) -> any:
    """ Позволяет выполнять неблокирующий вызов блокирующих функций. Например:
    await async_util_call(
        save_img_as, "ScreenShot.png", await page_instance.MakeScreenshot()
    )
    """
    return await asyncio.get_running_loop().run_in_executor(
        None, function, *args
    )


def find_instances(for_port: Optional[int] = None, browser: str = "chrome") -> Dict[int, int]:
    """ !!! ВНИМАНИЕ !!! На Windows 11 может быть отключен компонент WMI("Windows Management
    Instrumentation"), его нужно либо включить в разделе “Программы и компоненты” панели
    управления, либо установить из официальных источников.

    Используется для обнаружения уже запущенных браузеров в режиме отладки.
    Например:
            if browser_instances := find_instances():
                port, pid = [(k, v) for k, v in browser_instances.items()][0]
                browser_instance = Browser(debug_port=port, chrome_pid=pid)
            else:
                browser_instance = Browser()

            # Или для конкретного, известного порта:
            if browser_instances := find_instances(port):
                pid = browser_instances[port]
                browser_instance = Browser(debug_port=port, chrome_pid=pid)
            else:
                browser_instance = Browser()
    :param for_port:    - порт, для которого осуществляется поиск.
    :param browser:     - браузер, для которого запрашивается поиск.
    :return:            - словарь, ключами которого являются используемые порты запущенных
                            браузеров, а значениями, их ProcessID, или пустой словарь,
                            если ничего не найдено.
                            { 9222: 16017, 9223: 2001, ... }
    """
    result = {}
    if sys.platform == "win32":
        if "chrom" in browser: browser = "chrome.exe"
        elif "brave" in browser: browser = "brave.exe"
        elif "edge" in browser: browser = "msedge.exe"
        # else: ValueError("Not support browser: " + browser)
        cmd = f"WMIC PROCESS WHERE NAME='{browser}' GET Commandline,Processid"
        for line in subprocess.Popen(cmd, stdout=subprocess.PIPE).stdout:
            if b"--type=renderer" not in line and b"--remote-debugging-port=" in line:
                port, pid = re.findall(r"--remote-debugging-port=(\d+).*?(\d+)\s*$", line.decode())[0]
                port, pid = int(port), int(pid)
                if for_port == port: return {port: pid}
                result[port] = pid

    elif sys.platform == "linux":
        if "chrom" in browser: browser = "chrome"
        elif "brave" in browser: browser = "brave"
        elif "edge" in browser: browser = "msedge"
        # else: ValueError("Not support browser: " + browser)
        try: itr = map(int, subprocess.check_output(["pidof", browser]).split())
        except subprocess.CalledProcessError: itr = []
        for pid in itr:
            with open("/proc/" + str(pid) + "/cmdline") as f: cmd_line =  f.read()[:-1]
            if "--type=renderer" not in cmd_line and "--remote-debugging-port=" in cmd_line:
                port = int(re.findall(r"--remote-debugging-port=(\d+)", cmd_line)[0])
                if for_port == port: return {port: pid}
                result[port] = pid
    else:
        raise OSError(f"Platform '{sys.platform}' — not supported")
    return {} if for_port else result



from dataclasses import dataclass
from typing import Optional


@dataclass
class RemoteLocation:
    host: str; port: int


@dataclass
class TargetInfo:
    targetId: str; type: str; title: str; url: str; attached: bool
    openerId:         Optional[str] = None
    canAccessOpener: Optional[bool] = None
    openerFrameId:    Optional[str] = None
    browserContextId: Optional[str] = None



from typing import List
from .types import ProcessInfo, SystemData, GPUInfo


class SystemInfo:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/SystemInfo/
    """
    __slots__ = ("_connection",)

    def __init__(self, conn) -> None:
        from ...connection import Connection

        self._connection: Connection = conn

    async def getSystemInfo(self) -> SystemData:
        """
        Возвращает информацию о системе.
        https://chromedevtools.github.io/devtools-protocol/tot/SystemInfo/#method-getInfo
        :return:
        """
        result = await self._connection.call("SystemInfo.getInfo")
        return SystemData(
            gpu=GPUInfo(**result["gpu"]), modelName=result["modelName"],
            modelVersion=result["modelVersion"], commandLine=result["commandLine"]
        )

    async def getProcessInfo (self) -> List[ProcessInfo]:
        """
        Возвращает информацию обо всех запущенных в системе процессах.
        https://chromedevtools.github.io/devtools-protocol/tot/SystemInfo/#method-getProcessInfo
        :return:
        """
        result = await self._connection.call("SystemInfo.getProcessInfo")
        return [ProcessInfo(**i) for i in result]



import asyncio
from urllib.error import URLError

try:
    import ujson as json
    HAS_UJSON = True
except ModuleNotFoundError:
    HAS_UJSON = False
    import json
import warnings
import re, os, sys, signal, subprocess
from urllib.parse import quote
from os.path import expanduser
from inspect import iscoroutinefunction
from typing import List, Dict, Union, Optional, Tuple, Literal
from collections.abc import Sequence
from enum import Enum
from .connection import Connection
from .data import TargetConnectionInfo, TargetConnectionType, CommonCallback
from .exceptions import FlagArgumentContainError, NoTargetWithGivenIdFound
from .utils import get_request, find_browser_executable_path, log, async_util_call


class Browser:

    def __init__(
            self,
            profile_path: str = "testProfile",
            dev_tool_profiles:  bool = False,
            url: Optional[Union[str, bytes]] = None,
            flags:  Optional["FlagBuilder"] = None,
            browser_path: str = "",
            debug_port:   Union[str, int] = 9222,
            browser_pid:  int = 0,
            app:         bool = False,
            browser_exe:  str = "chrome",
            proxy_address:str = "http://127.0.0.1",
            proxy_port:   Union[str, int] = "",
            verbose:     bool = False,
            position: Optional[Tuple[int, int]] = None,
            sizes:    Optional[Tuple[int, int]] = None,
            prevent_restore: bool = False,

    ) -> None:
        """
        Все параметры — не обязательны.
        ==============================================================================================

        :param profile_path:    Путь до каталога, в который будет сохранена сессия профиля.
                                    Если не передан, или указано название папки, браузер
                                    сам создаст папку, по месту вызывающего кода. Если указан
                                    несуществующий путь, он будет создан.

                                        [-!!!-] ВАЖНО [-!!!-] - для запуска в режиме "headless",
                                        в "profile_path" нужно передать пустую строку.

        :param dev_tool_profiles:   Если 'profile_path' указан как имя папки, а не путь,
                                    профиль с указанным именем будет создан/получен в домашней
                                    директории пользователя, из каталога 'DevTools_Profiles',
                                    если значение установлено в True. (Linux, Windows)

        :param url:             Адрес, которым будет инициирован старт браузера.
                                    Со значением по умолчанию, будет открыта пустая страница.

        :param flags:           Экземпляр `FlagBuilder()` напичканный `CMDFlags()`.
                                    https://peter.sh/experiments/chromium-command-line-switches/

        :param browser_path:    Путь до исполняемого файла браузера. Имеет приоритет над
                                    аргументом `browser_exe`.

        :param debug_port:      Используется порт по умолчанию 9222.

        :param browser_pid:     ProcessID браузера. Используется методом kill(). Если
                                    передано значение большее нуля, считается, что производится
                                    подключение к уже существующему экземпляру браузера. Чтобы
                                    найти запущенный браузер в режиме отладки, воспользуйтесь
                                    `find_instances()`.

        :param app:             Запускает браузер в окне без пользовательского интерфейса,
                                    вроде адресной строки, кнопок навигации и прочих атрибутов.
                                    Распространяется только на первую страницу браузера. Прочие
                                    страницы будут открываться в обычном виде и в отдельном
                                    от первого окне.

        :param browser_exe:     Имя исполняемого файла браузера, который будет автоматизирован.
                                    Например: chrome, brave. Если 'browser_path' окажется пустым,
                                    будет произведена попытка найти его в системе по этому имени.

        :param proxy_port:      Если установлено, браузер будет запущен с флагом 'proxy' и
                                    все его запросы будут перенаправляться на этот порт, на
                                    локальном хосте.

        :param verbose:         Печатать некие подробности процесса?

        :param position:        Кортеж с иксом и игреком, в которых откроется окно браузера.

        :param sizes:           Кортеж с длиной и шириной в которые будет установлено окно браузера.

        :param prevent_restore: Предотвращать восстановление предыдущей сессии после крашей.
        """
        if sys.platform not in ("win32", "linux"): raise OSError(f"Platform '{sys.platform}' — not supported")
        self.dev_tool_profiles = dev_tool_profiles if profile_path else False

        if verbose and not HAS_UJSON:
            log("Call 'python -m pip install ujson' for install", lvl="[- UJSON IS NOT INSTALLED -]")

        if self.dev_tool_profiles:
            self.profile_path = os.path.join(expanduser("~"), "DevTools_Profiles", profile_path)
        elif profile_path != "":
            self.profile_path = os.path.abspath(profile_path)
        else:
            self.profile_path = ""

        preferences_path = ""
        if self.profile_path:
            preferences_path = os.path.join(self.profile_path, "Default", "Preferences")

        self.first_run = self.profile_path == "" or not os.path.exists(self.profile_path)

        # ? Предотвращать восстановление предыдущей сессии
        if not self.first_run and prevent_restore and preferences_path and os.path.exists(preferences_path):
            READ_WRITE = 33206 if sys.platform == "win32" else 33152
            READ_ONLY  = 33060 if sys.platform == "win32" else 33024
            # print("file attr =", os.stat(preferences_path).st_mode)
            # ? НЕ только для чтения
            if os.stat(preferences_path).st_mode == READ_WRITE:
                with open(preferences_path, "r") as f:
                    preferences = f.read()
                # ? тип завершения
                if exit_type := re.search(r'"exit_type":"(\w+)"', preferences):
                    # ? НЕ нормальный
                    if exit_type.group(1) != "Normal":
                        result = re.sub(r'"exit_type":"\w+"', '"exit_type":"Normal"', preferences)
                        with open(preferences_path, "w") as f: f.write(result)
                # ? Только для чтения
                # os.chmod(preferences_path, READ_ONLY)
                if verbose: log("Файл настроек — изменён и сохранён")
            else:
                # os.chmod(preferences_path, READ_WRITE)
                if verbose: log("Файл настроек — только для чтения")

        if browser_exe == "chrome":
            self.browser_name = "chrome"
            browser_exe = "chrome" if sys.platform == "win32" else "google-chrome"
        elif browser_exe == "brave":
            self.browser_name = "brave"
            browser_exe = "brave" if sys.platform == "win32" else "brave-browser"
        elif browser_exe == "chromium":
            self.browser_name = "chrome"
            browser_exe = "chromium" if sys.platform == "win32" else "chromium-browser"
        elif "edge" in browser_exe:
            self.browser_name = "edge"
            browser_exe = "msedge" if sys.platform == "win32" else "microsoft-edge"

        # ? Константы URL соответствующих вкладок
        self.NEW_TAB:       str = self.browser_name + "://newtab/"          # дефолтная вкладка
        self.SETTINGS:      str = self.browser_name + "://settings/"        # настройки
        self.BRAVE_REWARDS: str = self.browser_name + "://rewards/"         # вознаграждения (brave only)
        self.HISTORY:       str = self.browser_name + "://history/"         # история переходов
        self.BOOKMARKS:     str = self.browser_name + "://bookmarks/"       # закладки
        self.DOWNLOADS:     str = self.browser_name + "://downloads/"       # загрузки
        self.WALLET:        str = self.browser_name + "://wallet/"          # кошельки (brave only)
        self.EXTENSIONS:    str = self.browser_name + "://extensions/"      # расширения
        self.FLAGS:         str = self.browser_name + "://flags/"           # экспериментальные технологии

        if sys.platform == "win32":
            browser_path = browser_path if browser_path else find_browser_executable_path(browser_exe)
        else:   #  ! sys.platform == "linux"
            browser_path = browser_path if browser_path else os.popen("which " + browser_exe).read().strip()

        if not os.path.exists(browser_path) or not os.path.isfile(browser_path):
            raise FileNotFoundError(f"Переданный 'browser_path' => '{browser_path}' — не существует, или содержит ошибку")
        self.browser_path = browser_path

        if int(debug_port) <= 0:
            raise ValueError(f"Значение 'debug_port' => '{debug_port}' — должно быть положительным целым числом!")
        self.debug_port = str(debug_port)
        self.proxy_addres = proxy_address
        self.proxy_port = str(proxy_port)
        self.verbose = verbose

        # https://stackoverflow.com/questions/2381241/what-is-the-subprocess-popen-max-length-of-the-args-parameter
        # data_url_len_is_high = len(url[0]) > 32_767
        data_url_len = len(url) if url else 0
        # print("url =", url)
        if data_url_len > 30_000:
            warnings.warn(f"Length data url ({data_url_len}) is approaching to critical length = 32767 symbols!")

        b_name_len = len(self.browser_name)
        # Если "app" == True:
        _url_ = (
            # открыть dataURL без содержимого, если в "url" ничего не передано
            "--app=data:text/html," if url is None else
                # иначе установить переданную разметку, если она пришла как строка
                # и начало этой строки не содержит признаков url-адреса
                "--app=data:text/html," + quote(url)
                    if type(url) is str and "http" != url[:4] and self.browser_name != url[:b_name_len] else "--app=" + url
                        # передать "как есть", раз это строка содержащая url
                        if type(url) is str and "http" == url[:4] or self.browser_name == url[:b_name_len] else
                            # иначе декодировать и установить её как Base64
                            "--app=data:text/html;Base64," + url.decode()

            # иначе, открыть пустую страницу, если в "url" ничего не передано
        ) if app else "about:blank" if url is None else (
            # иначе передать разметку как data-url, если начало этой строки
            # не содержит признаков url-адреса
            "data:text/html," + quote(url)
                if type(url) is str and "http" != url[:4] and self.browser_name != url[:b_name_len] else url
                    # или передать "как есть", раз это строка содержащая url
                    if type(url) is str and "http" == url[:4] or self.browser_name == url[:b_name_len] else
                        # иначе декодировать и установить её как Base64
                        "data:text/html;Base64," + url.decode()
        )

        self.is_headless_mode = False
        self.browser_pid = browser_pid if browser_pid > 0 else self._run_browser(_url_, flags, position, sizes)

    def _run_browser(self, url: str, flags: "FlagBuilder", position: Optional[Tuple[int, int]] = None,
                     sizes: Optional[Tuple[int, int]] = None) -> int:
        """
        Запускает браузер с переданными флагами.
        :param url:             Адрес. Если передан, будет загружен в первой вкладке.
        :param flags:           Флаги
        :return:                ProcessID запущенного браузера
        """
        flag_box = FlagBuilder()
        flag_box.set(
            (CMDFlags.Common.remote_debugging_port, [self.debug_port]),
            (CMDFlags.Common.no_first_run, []),
            (CMDFlags.Common.no_default_browser_check, []),
            (CMDFlags.Test.log_file, ["null"]),
            (CMDFlags.Background.disable_breakpad, []),
            (CMDFlags.Background.no_recovery_component, []),
            (CMDFlags.Background.disable_sync, []),
            (CMDFlags.Background.disable_domain_reliability, []),
            (CMDFlags.Background.no_service_autorun, []),
            (CMDFlags.Performance.disable_gpu_shader_disk_cache, []),
            (CMDFlags.Performance.disk_cache_dir, ["null"]),
            (CMDFlags.Performance.media_cache_dir, ["null"]),
            (CMDFlags.Render.enable_features_WebUIDarkMode, []),
            (CMDFlags.Render.force_dark_mode, []),
        )
        run_args = [ self.browser_path ]
        flag_box.custom(url)

        # ! Default mode
        if self.profile_path:
            flag_box.add(CMDFlags.Common.user_data_dir, self.profile_path)
            if position is not None:
                flag_box.add(CMDFlags.Screen.window_position, *position)
            if sizes is not None:
                flag_box.add(CMDFlags.Screen.window_size, *sizes)

        # ! Headless mode
        else:
            flag_box.add(CMDFlags.Headless.headless)
            self.is_headless_mode = True

        if self.proxy_port:
            flag_box.add(CMDFlags.Other.proxy_server, self.proxy_addres + ":" + self.proxy_port)
            if self.verbose:
                log(f"Run browser {self.browser_name!r} on port: {self.debug_port} with proxy " +
                      f"on http://127.0.0.1:{self.proxy_port}")
        else:
            if self.verbose:
                log(f"Run browser {self.browser_name!r} on port: {self.debug_port}")

        if flags is not None:
            flag_box += flags

        run_args += flag_box.flags()

        pipe = subprocess.PIPE if not self.verbose else None
        return subprocess.Popen(run_args, stdout=pipe, stderr=pipe).pid

    def kill(self) -> None:
        """  Убивает процесс браузера. """
        try:
            os.kill(self.browser_pid, signal.SIGTERM)
        except PermissionError:
            pass

    async def getAllTargetsConnectionInfo(self) -> List[TargetConnectionInfo]:
        """ Возвращает список описаний соединений со всеми
        существующими типами соединений.
        """
        return [TargetConnectionInfo(**i) for i in await self.getConnectionList()]

    async def getConnectionsByType(
            self, conn_type: Union[str, TargetConnectionType]) -> List[TargetConnectionInfo]:
        """ Возвращает список описаний соединений, удовлетворяющих
        запрошенному типу соединения.
        :param conn_type:     Тип соединения. Например: "page"
        """
        t = conn_type if type(conn_type) is str else conn_type.value
        return [ti for ti in await self.getAllTargetsConnectionInfo() if ti.type == t]

    async def getConnectionList(self) -> List[dict]:
        """
        Запрашивает у браузера список всех его дочерних процессов,
        включая табы(вкладки/страницы), воркеры, расширения, сервисы и прочее.

        :return: [ {
                    "description": "",
                    "devtoolsFrontendUrl": "/devtools/inspector.html?ws=localhost:9222/devtools/page/DAB7FB6187B554E10B0BD18821265734",
                    "id": "DAB7FB6187B554E10B0BD18821265734",
                    "title": "Yahoo",
                    "type": "page",
                    "url": "https://www.yahoo.com/",
                    "webSocketDebuggerUrl": "ws://localhost:9222/devtools/page/DAB7FB6187B554E10B0BD18821265734"
                }, { ... } ]
        """
        result = await async_util_call(
            get_request, f"http://127.0.0.1:{self.debug_port}/json/list")

        if self.verbose: log("GetPageList() => " + result)
        return json.loads(result)

    async def getConnectionBy(
            self, key: Union[str, int],
            value: Union[str, int],
            match_mode: Literal["exact", "contains", "startswith"] = "exact",
            index: int = 0,
            callback: CommonCallback = None) -> Optional[Connection]:
        """
        Организует выбор нужного соединения из процессов браузера по следующим критериям:
        :param key:                 По ключу из словаря. Список ключей смотри
                                        в возвращаемых значениях GetPageList()
        :param value:               Значение ключа, которому должен соответствовать выбор
        :param match_mode:          Значение ключа:
                                        "exact"      - полностью совпадает с value,
                                        "contains"   - содержит value,
                                        "startswith" - начинается с value
        :param index:               Поскольку открытых страниц с одинаковым ключом может
                                        быть несколько, предоставляется возможность выбрать
                                        по индексу. По умолчанию = 0
        :param callback:            Корутина, которой будет передаваться контекст абсолютно
                                        всех событий страницы в виде словаря. Если передан,
                                        включает уведомления домена "Runtime" для общения
                                        со страницей.

        :return:        <Connection>
        """

        if callback is not None and not iscoroutinefunction(callback):
            raise TypeError("Argument 'callback' must be a coroutine")

        counter = 0; v = value.lower()
        for page_data in await self.getConnectionList():
            data = page_data[key]
            if ((match_mode == "exact" and data == v)
                or (match_mode == "contains" and data.find(v) > -1 )
                    or (match_mode == "startswith" and data.find(v) == 0)):
                if counter == index:
                    conn = Connection(
                        page_data["webSocketDebuggerUrl"],
                        page_data["id"],
                        page_data["devtoolsFrontendUrl"],
                        callback,
                        self.profile_path == "",
                        self.verbose,
                        self.browser_name
                    )

                    await conn.activate()
                    return conn
                counter += 1
        return None

    async def getConnection(
            self, index: int = 0,
            conn_type: str = "page",
            callback: CommonCallback = None) -> Connection:
        """
        Получает страницу браузера по индексу. По умолчанию, первую.
        :param index:       - Желаемый индекс страницы начиная с нуля.
        :param conn_type:   - Тип "page" | 'background_page' | 'service_worker' | ???
        :param callback:    - Корутина, которой будет передаваться контекст абсолютно
                                всех событий страницы в виде словаря.
        :return:        <Connection>
        """
        return await self.getConnectionBy("type", conn_type, "exact", index, callback)

    async def getConnectionByID(
            self, conn_id: str,
            callback: CommonCallback = None) -> Connection:
        """
        Получает страницу браузера по уникальному идентификатору.
        :param conn_id:     - Желаемый идентификатор страницы. Он же 'targetId'.
        :param callback:    - Корутина, которой будет передаваться контекст абсолютно
                                всех событий страницы в виде словаря.
        :return:        <Connection>
        """
        return await self.getConnectionBy("id", conn_id, "exact", 0, callback)

    async def getConnectionByTitle(
            self, value: str,
            match_mode: Literal["exact", "contains", "startswith"] = "startswith",
            index: int = 0,
            callback: CommonCallback = None) -> Connection:
        """
        Получает страницу браузера по заголовку. По умолчанию, первую.
        :param value:       - Текст, который будет сопоставляться при поиске.
        :param match_mode:  - Тип сопоставления(по умолчанию 'startswith').
                                Может быть только:
                                    * exact      - полное соответствие заголовка и value
                                    * contains   - заголовок содержит value
                                    * startswith - заголовок начинается с value
        :param index:       - Желаемый индекс страницы начиная с нуля.
        :param callback:    - Корутина, которой будет передаваться контекст абсолютно
                                всех событий страницы в виде словаря.
        :return:        <Connection>
        """
        return await self.getConnectionBy("title", value, match_mode, index, callback)

    async def getConnectionByURL(
            self, value: str,
            match_mode: Literal["exact", "contains", "startswith"] = "startswith",
            index: int = 0,
            callback: CommonCallback = None) -> Connection:
        """
        Получает страницу браузера по её URL. По умолчанию, первую.
        :param value:       - Текст, который будет сопоставляться при поиске.
        :param match_mode:  - Тип сопоставления(по умолчанию 'startswith').
                                Может быть только:
                                    * exact      - полное соответствие URL и value
                                    * contains   - URL содержит value
                                    * startswith - URL начинается с value
        :param index:       - Желаемый индекс страницы начиная с нуля.
        :param callback:    - Корутина, которой будет передаваться контекст абсолютно
                                всех событий страницы в виде словаря.
        :return:        <Connection>
        """
        return await self.getConnectionBy("url", value, match_mode, index, callback)

    async def createTab(
            self, url: str = "about:blank",
            newWindow: bool = False,
            background: bool = False,
            wait_for_create: bool = True,
            callback: CommonCallback = None) -> Optional[Connection]:
        """
        Создаёт новую вкладку в браузере.
        :param url:                     - (optional) Адрес будет открыт при создании.
        :param newWindow:               - (optional) Если 'True' — страница будет открыта в новом окне.
        :param background:              - (optional) Если 'True' — страница будет открыта в фоне.
        :return:                    * <Connection>
        """
        while not (tmp := await self.getConnection(callback=callback)):
            await asyncio.sleep(.5)
        page_id = await tmp.Target.createTarget(url, newWindow=newWindow, background=background)
        if wait_for_create:
            while not (page := await self.getConnectionByID(page_id)):
                await asyncio.sleep(.5)
        else:
            page = await self.getConnectionByID(page_id)
        return page
    
    async def showInspector(
            self, conn: Connection,
            new_window: bool = True,
            callback: CommonCallback = None) -> Optional[Connection]:
        """
        Открывает новую вкладку с дебаггером для инспектируемой страницы.
        :param conn:            - Инспектируемая страница. Может принадлежать любому браузеру.
        :param new_window:      - Создать target в отдельном окне?
        :return:        <Connection>
        """
        return await self.createTab(
            "http://127.0.0.1:" + str(self.debug_port) + conn.frontend_url, new_window, callback=callback
        )

    async def createPopupWindow(
            self, conn: Connection,
            url: str = "about:blank",
            callback: CommonCallback = None) -> Optional[Connection]:
        """ Создаёт всплывающее окно с минимумом интерфейса браузера".
        :param url:             - Адрес, ресурс которого будет загружен
        :param conn:            - Родительская страница, инициатор
        :return:        Connection or None
        """
        await conn.extend.injectJS(
            f'window.open("{url}", "blank_window_name", "popup,noopener,noreferrer");')
        return await self.getConnectionByOpener(conn, callback=callback)

    async def getConnectionByOpener(
            self, conn: Connection,
            callback: CommonCallback = None) -> Optional[Connection]:
        """ Возвращает последнее созданное соединение со страницей, открытие которого
        инициировано с конкретной страницы. Например, при использовании JavaScript
        "window.open()".
        :param conn:            - Родительская страница, инициатор
        :return:        Connection or None
        """
        for target_info in await conn.Target.getTargets():
            if target_info.openerId == conn.conn_id:
                return await self.getConnectionByID(target_info.targetId, callback=callback)
        return None

    async def getConnectionsByOpener(
            self, conn: Connection,
            callback: CommonCallback = None) -> List[Connection]:
        """ Возвращает список всех соединений, открытие которых инициировано с конкретной
        страницы. Например, при использовании JavaScript "window.open()".
        :param conn:            - Родительская страница, инициатор открытых окон
        :return:        List[Connection]
        """
        connections = []
        for target_info in await conn.Target.getTargets():
            if target_info.openerId == conn.conn_id:
                connections.append(await self.getConnectionByID(
                    target_info.targetId, callback=callback))
        return connections

    async def waitFirstTab(
            self, timeout: float = 20.0,
            callback: CommonCallback = None) -> Connection:
        """ Дожидается получения соединения или вызывает исключение
        'asyncio.exceptions.TimeoutError' по истечении таймаута.
        """
        return await asyncio.wait_for(self.getFirstTab(callback), timeout)

    async def getFirstTab(self, callback: CommonCallback = None) -> Connection:
        """
        Безусловно дожидается соединения со страницей.
        """
        while True:
            try:
                while (conn := await self.getConnection(callback=callback)) is None:
                    await asyncio.sleep(.5)
                return conn
            except URLError: await asyncio.sleep(1)

    async def close(self) -> None:
        """ Корректно закрывает браузер если остались ещё его инстансы """
        if conn := await self.getConnection():
            await conn.Browser.close()

    async def closeAllTabsExcept(self, *except_list: Connection) -> None:
        """ Закрывает все страницы браузера, кроме переданных. """
        for conn_info in await self.getAllTargetsConnectionInfo():
            if conn_info.type == "page":
                condition = False
                for conn in except_list:
                    condition |= conn.conn_id == conn_info.id
                if not condition:
                    i = 4
                    try:
                        while (tab := await self.getConnectionByID(conn_info.id)) is None and i:
                            await asyncio.sleep(.5)
                            i -= 1
                        if tab: await tab.Target.close()
                    except NoTargetWithGivenIdFound:
                        pass

    async def getFramesFor(self, conn: Connection) -> List[Connection]:
        """ Возвращает список iFrame для указанного соединения. """
        return [
            await self.getConnectionByID(data["id"])
            for data in await self.getConnectionList()
            if data["type"] == "iframe" and data["parentId"] == conn.conn_id
        ]

    def __eq__(self, other: "Browser") -> bool:
        return self.debug_port == other.debug_port

    def __hash__(self) -> int:
        return hash(self.debug_port)


class FlagBuilder:
    """ Обеспечивает последовательность неповторяющихся флагов
    для запуска браузера.
    """
    def __init__(self) -> None:
        self._flags: Dict[str, Tuple[str]] = {}

    def add(self, flag: "CMDFlag", *args: Union[str, int, float]) -> None:
        """ Принимает один флаг и его аргументы. """
        f: str = flag.value
        if f[-1] == "=":
            if not args:
                raise FlagArgumentContainError(
                    f"Для флага {flag.name} ожидается аргумент, или список аргументов.")

            self._flags[f] = tuple(map(str, args))
        else:
            self._flags[f] = tuple()

    def set(self, *flags: Tuple["CMDFlag", Sequence[Union[str, int, float]]]) -> None:
        """ Принимает произвольную последовательность флагов, из кортежей,
         содержащих сам флаг и его аргументы:
            object.set(
                (CMDFlags.Background.disable_breakpad, []),
                (CMDFlags.Background.no_recovery_component, [])
            )
        """
        for cmd_flag, args in flags:
            self.add(cmd_flag, *args)

    def custom(self, flag: str) -> None:
        """ Принимает строку в качестве флага.
        object.custom("--mute-audio")
        """
        self._flags[flag] = tuple()

    def flags(self) -> List[str]:
        """ Возвращает форматированный список флагов
        для запуска процесса браузера.
        """
        return [
            cmd_flag + ",".join(args)
            for cmd_flag, args in self._flags.items()
        ]

    def __str__(self) -> str:
        return "[\n" + ",\n\t".join(self.flags()) + "\n]"

    def __add__(self, other: "FlagBuilder") -> "FlagBuilder":
        if not isinstance(other, FlagBuilder):
            raise TypeError(f"Ожидаемый тип: {self.__class__.__name__}; "
                            f"полученный тип: {other.__class__.__name__}")

        flags = {}
        for cmd_flag, args in self._flags.items():
            flags[cmd_flag] = args
        for cmd_flag, args in other._flags.items():
            flags[cmd_flag] = args

        result = FlagBuilder()
        setattr(result, "_flags", flags)
        return result


class CMDFlag(Enum): pass

class CMDFlags:
    """https://github.com/GoogleChrome/chrome-launcher/blob/master/docs/chrome-flags-for-tools.md"""

    class Common(CMDFlag):
        # ? Commonly unwanted browser features
        # ! Порт подключения
        remote_debugging_port = "--remote-debugging-port="
        # ! Отключает обнаружение фишинга на стороне клиента.
        disable_client_side_phishing_detection = "--disable-client-side-phishing-detection"
        # ! Отключить все расширения Chrome.
        disable_extensions = "--disable-extensions"
        # ! Отключить некоторые встроенные расширения, на которые не влияет --disable-extensions.
        disable_component_extensions_with_background_pages = "--disable-component-extensions-with-background-pages"
        # ! Отключить установку приложений по умолчанию.
        disable_default_apps = "--disable-default-apps"
        # ! Беззвучный режим.
        mute_audio = "--mute-audio"
        # ! Отключить проверку браузера по умолчанию, не предлагать установить ее как таковую
        no_default_browser_check = "--no-default-browser-check"
        # ! Пропустить мастера первого запуска.
        no_first_run = "--no-first-run"
        # ! Используйте поддельное устройство для Media Stream, чтобы заменить камеру и микрофон.
        use_fake_device_for_media_stream = "--use-fake-device-for-media-stream"
        # ! Использовать файл для фальшивого захвата видео (.y4m или .mjpeg).--use-fake-device-for-media-stream
        # !     Принимает путь к файлу.
        use_file_for_fake_video_capture = "--use-file-for-fake-video-capture="  # * $
        # ! Отключает WebGL
        disable_webgl = "--disable-webgl"
        # ! Запускает браузер в режиме "Инкогнито"
        incognito = "--incognito"
        # ! Принимает путь к каталогу, где будет позиционирован весь кеш браузера, включая профили и прочее.
        user_data_dir = "--user-data-dir="      # * $

    class Performance(CMDFlag):
        # ? Performance & web platform behavior
        # ! Следующие два флага применяются вместе, отключая применение одного и того же происхождения (не рекомендуется).
        disable_web_security = "--disable-web-security"
        allow_running_insecure_content = "--allow-running-insecure-content"
        # --------------------------------
        # ! Отключить автоплей видео.
        autoplay_policy_user_gesture_required = "--autoplay-policy=user-gesture-required"
        # ! Отключить регулировку таймеров на фоновых страницах/вкладках.
        disable_background_timer_throttling = "--disable-background-timer-throttling"
        # ! Отключите фоновую визуализацию для закрытых окон. Сделано для тестов, чтобы избежать недетерминированного
        # !     поведения.
        disable_backgrounding_occluded_windows = "--disable-backgrounding-occluded-windows"
        # ! Отключает потоковую передачу сценариев V8.
        disable_features_ScriptStreaming = "--disable-features=ScriptStreaming"
        # ! Подавляет диалоги зависания монитора в процессах визуализации. Это может позволить обработчикам медленной
        # !     выгрузки на странице предотвратить закрытие вкладки, но в этом случае можно использовать диспетчер задач
        # !     для завершения вызывающего нарушение процесса.
        disable_hang_monitor = "--disable-hang-monitor"
        # ! Некоторые функции javascript можно использовать для заполнения процесса браузера IPC. По умолчанию защита
        # !     включена, чтобы ограничить количество отправляемых IPC до 10 в секунду на кадр. Этот флаг отключает его.
        disable_ipc_flooding_protection = "--disable-ipc-flooding-protection"
        # ! отключает веб-уведомления и API-интерфейсы Push.
        disable_notifications = "--disable-notifications"
        # ! отключить блокировку всплывающих окон.
        disable_popup_blocking = "--disable-popup-blocking"
        # ! перезагрузка страницы, полученной с помощью POST, обычно выводит пользователю запрос.
        disable_prompt_on_repost = "--disable-prompt-on-repost"
        # ! отключает вкладки, не находящиеся на переднем плане, от получения более низкого приоритета процесса. Это (само
        # !     по себе) не влияет на таймеры или поведение рисования.
        disable_renderer_backgrounding = "--disable-renderer-backgrounding"
        # ! https://stackoverflow.com/questions/59462637/list-of-js-flags-that-can-be-used-for-chrome-on-windows
        js_flags = "--js-flags="                # * $
        # ! Отключает шейдеры GPU в кеше на диске.
        disable_gpu_shader_disk_cache = "--disable-gpu-shader-disk-cache"
        # ! Принимает путь расположения кэша на диске, отличный от UserDatadir. Отключить: '--disk-cache-dir=null'
        disk_cache_dir = "--disk-cache-dir="    # * $
        # ! Задает максимальное дисковое пространство, используемое дисковым кешем, в байтах.
        disk_cache_size = "--disk-cache-size="  # * $
        # ! Принимает путь расположения media-кэша на диске, отличный от UserDatadir. Отключить: '--media-cache-dir=null'
        media_cache_dir = "--media-cache-dir="    # * $
        # ! Задает максимальное дисковое пространство, используемое дисковым кешем для медиа, в байтах.
        media_cache_size = "--media-cache-size="  # * $

    class Test(CMDFlag):
        # ? Test & debugging flags
        # ! Сообщает, выполняются ли в коде тесты браузера (это изменяет URL-адрес запуска, используемый оболочкой
        # !     содержимого, а также отключает функции, которые могут сделать тесты ненадежными [например, мониторинг
        # !     нехватки памяти]).
        browser_test = "--browser-test"
        # ! Избегает сообщений типа «Новый принтер в вашей сети».
        disable_device_discovery_notifications = "--disable-device-discovery-notifications"
        # ! Отключает несколько вещей, которые не подходят для автоматизации:
        # !     * отключает всплывающие уведомления о запущенных разработках/распакованных расширениях
        # !     * отключает пользовательский интерфейс сохранения пароля (который охватывает вариант использования
        # !         удаленного --disable-save-password-bubble флага)
        # !     * отключает анимацию информационной панели
        # !     * отключает автоперезагрузку при сетевых ошибках
        # !     * означает, что запрос проверки браузера по умолчанию не отображается
        # !     * избегает отображения этих 3 информационных панелей: ShowBadFlagsPrompt, GoogleApiKeysInfoBarDelegate,
        # !         ObsoleteSystemInfoBarDelegate
        # !     * добавляет эту информационную панель:
        # !         https://user-images.githubusercontent.com/39191/30349667-92a7a086-97c8-11e7-86b2-1365e3d407e3.png
        enable_automation = "--enable-automation"
        # ! Включает ведения журнала. Больше подходит для процесса серверного типа.
        enable_logging_stderr = "--enable-logging=stderr"
        # ! 0 — означает INFO и выше
        log_level = "--log-level="              # * $
        # ! Переопределяет дефолтное имя файла лога. Чтобы отключить: '--log-file=null'
        log_file = "--log-file="                # * $
        # ! Избегает потенциальной нестабильности при использовании Gnome Keyring или кошелька KDE.
        password_store_basic = "--password-store=basic"
        # ! Более безопасно, чем использование протокола через веб-сокет
        remote_debugging_pipe = "--remote-debugging-pipe"
        # ! Информационная панель не отображается, когда расширение Chrome подключается к странице с помощью
        # !     chrome.debuggerpage. Требуется для прикрепления к фоновым страницам расширения.
        silent_debugger_extension_api = "--silent-debugger-extension-api"
        # ! Похож на флаг --enable-automation
        # !     * позволяет избежать создания заглушек приложений в ~/Applications на Mac.
        # !     * делает коды выхода немного более правильными
        # !     * списки переходов для навигации в Windows не обновляются
        # !     * не запускается какой-либо хром StartPageService
        # !     * отключает инициализацию службы chromecast
        # !     * расширения компонентов с фоновыми страницами не включаются во время тестов, потому что они создают много
        # !         фонового поведения, которое может мешать
        # !     * при выходе из браузера он отключает дополнительные проверки, которые могут остановить этот процесс выхода.
        # !         (например, несохраненные изменения формы или необработанные уведомления профиля..)
        test_type = "--test-type"

    class Background(CMDFlag):
        # ? Background updates, networking, reporting
        # ! Отключить различные фоновые сетевые службы, включая обновление расширений, службу безопасного просмотра,
        # !     детектор обновлений, перевод, UMA.
        disable_background_networking = "--disable-background-networking"
        # ! Отключает сбор аварийных дампов (отчеты уже отключены в Chromium)
        disable_breakpad = "--disable-breakpad"
        # ! Не обновлять «компоненты» браузера, перечисленные в chrome://components/
        disable_component_update = "--disable-component-update"
        # ! Отключает мониторинг надежности домена, который отслеживает, возникают ли у браузера проблемы с доступом к
        # !     сайтам, принадлежащим Google, и загружает отчеты в Google.
        disable_domain_reliability = "--disable-domain-reliability"
        # ! Отключает синхронизацию с учетной записью Google.
        disable_sync = "--disable-sync"
        # ! Используется для включения отчетов о сбоях Breakpad в среде отладки, где отчеты о сбоях обычно скомпилированы,
        # !     но отключены.
        enable_crash_reporter_for_testing = "--enable-crash-reporter-for-testing"
        # ! Отключить отправку отчетов в UMA, но разрешить сбор.
        metrics_recording_only = "--metrics-recording-only"
        # ! Предотвращает загрузку и выполнение компонентов восстановления.
        no_recovery_component = "--no-recovery-component"
        # ! Запрещает процессам-сервисам добавлять себя в автозапуск. Это не удаляет существующие регистрации
        # !    автозапуска, а просто предотвращает регистрацию новых служб.
        no_service_autorun = "--no-service-autorun"

    class Render(CMDFlag):
        # ? Rendering & GPU
        # ! экспериментальный мета-флаг. Устанавливает приведенные ниже 7 флагов, переводящие браузер в режим, в котором
        # !     рендеринг (радиус границы и т. д.) является детерминированным, а начальные кадры должны выдаваться по
        # !     протоколу DevTools.
        deterministic_mode = "--deterministic-mode"

        run_all_compositor_stages_before_draw = "--run-all-compositor-stages-before-draw"
        disable_new_content_rendering_timeout = "--disable-new-content-rendering-timeout"
        enable_begin_frame_control = "--enable-begin-frame-control"
        disable_threaded_animation = "--disable-threaded-animation"
        disable_threaded_scrolling = "--disable-threaded-scrolling"
        disable_checker_imaging = "--disable-checker-imaging"
        disable_image_animation_resync = "--disable-image-animation-resync"
        # -------------
        # ! не откладывать отрисовку коммитов (обычно используется, чтобы избежать мигания нестилизованного содержимого).
        disable_features_PaintHolding = "--disable-features=PaintHolding"
        disable_partial_raster = "--disable-partial-raster"
        # ! Не используйте оптимизацию высокопроизводительного ЦП, обнаруженную во время выполнения, в Skia.
        disable_skia_runtime_opts = "--disable-skia-runtime-opts"
        # ! Экономит немного памяти, перемещая процесс графического процессора в поток процесса браузера.
        in_process_gpu = "--in-process-gpu"
        # ! выберите, какую реализацию GL должен использовать процесс GPU. Возможные варианты: desktop — любой рабочий стол
        # !     OpenGL, установленный пользователем (по умолчанию для Linux и Mac). egl — любой EGL / GLES2, который
        # !     пользователь установил (по умолчанию для Windows - фактически ANGLE). swiftshader — Программный рендерер
        # !     SwiftShader.
        use_gl = "--use-gl="    # * $
        # + DARK_MODE -- требует двух следующих флагов
        # ! Включает тёмный режим в браузере.
        enable_features_WebUIDarkMode = "--enable-features=WebUIDarkMode"
        # ! Принуждает использовать темный режим в пользовательском интерфейсе для платформ, которые его поддерживают.
        force_dark_mode = "--force-dark-mode"



    class Screen(CMDFlag):
        # ? Window & screen management
        # ! Запускает браузер в режиме KIOSK
        kiosk = "--kiosk"
        # ! Запрещает браузеру resize и убирает некоторые его элементы.
        force_app_mode = "--force-app-mode"
        # ! Все всплывающие окна и вызовы window.open завершатся ошибкой.
        block_new_web_contents = "--block-new-web-contents"
        # ! Заставить все мониторы обрабатываться так, как если бы они имели указанный цветовой профиль.
        force_color_profile_SRGB = "--force-color-profile=srgb"
        # ! Переопределяет коэффициент масштабирования устройства для пользовательского интерфейса браузера и содержимого.
        # !     int or float
        force_device_scale_factor = "--force-device-scale-factor="
        # ! Каждая ссылка запускается в новом окне.
        new_window = '--new-window'
        # ! Принимает X и Y позицию верхнего левого угла, окна браузера.
        window_position="--window-position="            # * $
        # ! Принимает ширину и высоту окна браузера.
        window_size = "--window-size="                  # * $

    class Process(CMDFlag):
        # ? Process management
        # ! Отключает OOPIF. https://www.chromium.org/Home/chromium-security/site-isolation
        disable_features_site_per_process = "--disable-features=site-per-process"
        # ! Запускает визуализатор и плагины в том же процессе, что и браузер.
        single_process = "--single-process"

    class Headless(CMDFlag):
        # ? Headless
        # ! Headless
        headless = "--headless"
        # ! Часто используется в сценариях Lambda, Cloud Functions.
        disable_dev_shm_usage = "--disable-dev-shm-usage"
        # ! Запуск без песочницы. НЕ РЕКОМЕНДУЕТСЯ!!!
        no_sandbox = "--no-sandbox"
        # ! С 2021 года не требуется.
        disable_gpu = "--disable-gpu"

    class Other(CMDFlag):
        """ https://niek.github.io/chrome-features/ """
        # ? Принимают любое кол-во аргументов
        enable_features = "--enable-features="      # * $
        disable_features = "--disable-features="    # * $
        # ! Принимает адрес и порт прокси:
        # !     http://192.168.0.1:2233
        proxy_server = "--proxy-server="            # * $



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union


class ConnectionType(Enum):
    none = "none"
    cellular2g = "cellular2g"
    cellular3g = "cellular3g"
    cellular4g = "cellular4g"
    bluetooth = "bluetooth"
    ethernet = "ethernet"
    wifi = "wifi"
    wimax = "wimax"
    other = "other"


@dataclass
class LoadNetworkResourcePageResult:
    success: bool
    netError: Optional[int] = None
    netErrorName: Optional[str] = None
    httpStatusCode: Optional[int] = None
    stream: Optional[str] = None                # ! IO.StreamHandle
    headers: Optional[dict] = None


@dataclass
class Request:
    url: str
    method: str
    headers: dict
    initialPriority: str                        # ! Allowed Values: VeryLow, Low, Medium, High, VeryHigh
    referrerPolicy: str                         # ! Allowed Values: unsafe-url, no-referrer-when-downgrade,
                                                # !     no-referrer, origin, origin-when-cross-origin, same-origin,
                                                # !     strict-origin, strict-origin-when-cross-origin
    trustTokenParams: Optional["TrustTokenParams"]
    postDataEntries: Optional[list["PostDataEntry"]]
    urlFragment: Optional[str] = None
    postData: Optional[str] = None
    hasPostData: Optional[bool] = None          # ! true if postData is present
    _postDataEntries: Optional[list["PostDataEntry"]] = field(init=False, repr=False, default=None)
    mixedContentType: Optional[str] = None      # ! Allowed Values: blockable, optionally-blockable, none
    isLinkPreload: Optional[bool] = None
    _trustTokenParams: Optional["TrustTokenParams"] = field(init=False, repr=False, default=None)
    isSameSite: Optional[bool] = None

    @property
    def postDataEntries(self) -> list["PostDataEntry"]:
        return self._postDataEntries

    @postDataEntries.setter
    def postDataEntries(self, data: list[dict[str, Union[str, None]]]) -> None:
        if not isinstance(data, property):
            self._postDataEntries = [PostDataEntry(**item) for item in data]
        else:
            self._postDataEntries = None

    @property
    def trustTokenParams(self) -> "TrustTokenParams":
        return self._trustTokenParams

    @trustTokenParams.setter
    def trustTokenParams(self, data: dict) -> None:
        self._trustTokenParams = TrustTokenParams(**data) if not isinstance(data, property) else None


@dataclass
class PostDataEntry:
    bytes: Optional[str] = None


@dataclass
class TrustTokenParams:
    type: str                                   # ! Allowed Values: Issuance, Redemption, Signing
    refreshPolicy: str                          # ! Allowed Values: UseCached, Refresh
    issuers: Optional[list[str]] = None



@dataclass
class Cookie:
    name: str
    value: str
    domain: str
    path: str
    expires: float
    size: int
    httpOnly: bool
    secure: bool
    session: bool
    priority: str                       # Allowed Values: Low, Medium, High
    sameParty: bool
    sourceScheme: str                   # Allowed Values: Unset, NonSecure, Secure
    sourcePort: int                     # Valid values are {-1, [1, 65535]}, -1 indicates an unspecified port.
    sameSite: Optional[str] = None      # Allowed status Values: Strict, Lax, None
    partitionKey: Optional[str] = None
    partitionKeyOpaque: Optional[bool] = None
